from __future__ import absolute_import, division, print_function # confidence high

import copy, os

from astropy.io import fits
import numpy as N
from math import *

from . import fileutil

# Convenience definitions...
yes = True
no = False

DEGTORAD = fileutil.DEGTORAD
RADTODEG = fileutil.RADTODEG
DIVMOD = fileutil.DIVMOD
DEFAULT_PREFIX = 'O'

#
# History
#
# 30-Mar-2002 WJH: Added separate XYtoSky interface function.
# 19-Apr-2002 WJH: Corrected 'ddtohms' for error in converting neg. dec.
# 20-Sept-2002 WJH: replaced all references to 'keypar' with calls to 'hselect'
#                   This avoids any parameter writes in the pipeline.
# 03-Dec-2002 WJH: Added 'new' parameter to WCSObject to make creating an
#                   object from scratch unambiguous and free from filename
#                   collisions with user files.
# 23-Apr-2003 WJH: Enhanced to search entire file for header with WCS keywords
#                   if no extension was specified with filename.
# 6-Oct-2003 WJH:  Modified to use the 'fileutil.getHeader' function or
#                   accept a PyFITS/readgeis header object.  Removed
#                   any explicit check on whether the image was FITS or
#                   not.
# 5-Feb-2004 WJH:  Added 'recenter' method to rigorously shift the WCS from
#                   an off-center reference pixel position to the frame center
#
# 24-Jan-2005 WJH: Added methods and attributes for working with archived
#                   versions of WCS keywords.  Archived keywords will be
#                   treated as 'read-only' if they already exist, unless
#                   specifically overwritten.
#
# 30-Mar-2005 WJH: 'read_archive' needed to be modified to use existing prefix
#                   found in header, if one exists, for computing archive pscale.
#
# 20-Jun-2005 WJH: Support for constant-value arrays using NPIX/PIXVALUE added to
#                   class.  The output reference WCS now creates a constant-value
#                   array for the extension as well in order to be FITS compliant.
#                   WCS keywords now get written out in a set order to be FITS compliant.
#                   New method, get_orient, added to always allow access to computed
#                   orientation regardless of orientat keyword value.
#
# 29-June-2005 WJH: Multiple WCS extensions are not created when running
#                   'createReferenceWCS'.
#




__version__ = '1.2.3 (11-Feb-2011)'

def help():
    print('wcsutil Version '+str(__version__)+':\n')
    print(WCSObject.__doc__)
#################
#
#
#               Coordinate Transformation Functions
#
#
#################


def ddtohms(xsky,ysky,verbose=no):

    """ Convert sky position(s) from decimal degrees to HMS format."""

    xskyh = xsky /15.
    xskym = (xskyh - N.floor(xskyh)) * 60.
    xskys = (xskym - N.floor(xskym)) * 60.

    yskym = (N.abs(ysky) - N.floor(N.abs(ysky))) * 60.
    yskys = (yskym - N.floor(yskym)) * 60.

    if isinstance(xskyh,N.ndarray):
        rah,dech = [],[]
        for i in range(len(xskyh)):
            rastr = repr(int(xskyh[i]))+':'+repr(int(xskym[i]))+':'+repr(xskys[i])
            decstr = repr(int(ysky[i]))+':'+repr(int(yskym[i]))+':'+repr(yskys[i])
            rah.append(rastr)
            dech.append(decstr)
            if verbose:
                print('RA = ',rastr,', Dec = ',decstr)
    else:
        rastr = repr(int(xskyh))+':'+repr(int(xskym))+':'+repr(xskys)
        decstr = repr(int(ysky))+':'+repr(int(yskym))+':'+repr(yskys)
        rah = rastr
        dech = decstr
        if verbose:
            print('RA = ',rastr,', Dec = ',decstr)

    return rah,dech


def troll(roll, dec, v2, v3):
    """ Computes the roll angle at the target position based on::

            the roll angle at the V1 axis(roll),
            the dec of the target(dec), and
            the V2/V3 position of the aperture (v2,v3) in arcseconds.

        Based on the algorithm provided by Colin Cox that is used in
        Generic Conversion at STScI.
    """
    # Convert all angles to radians
    _roll = DEGTORAD(roll)
    _dec = DEGTORAD(dec)
    _v2 = DEGTORAD(v2 / 3600.)
    _v3 = DEGTORAD(v3 / 3600.)

    # compute components
    sin_rho = sqrt((pow(sin(_v2),2)+pow(sin(_v3),2)) - (pow(sin(_v2),2)*pow(sin(_v3),2)))
    rho = asin(sin_rho)
    beta = asin(sin(_v3)/sin_rho)
    if _v2 < 0: beta = pi - beta
    gamma = asin(sin(_v2)/sin_rho)
    if _v3 < 0: gamma = pi - gamma
    A = pi/2. + _roll - beta
    B = atan2( sin(A)*cos(_dec), (sin(_dec)*sin_rho - cos(_dec)*cos(rho)*cos(A)))

    # compute final value
    troll = RADTODEG(pi - (gamma+B))

    return troll

#################
#
#
#               Coordinate System Class
#
#
#################

class WCSObject:
    """ This class should contain the WCS information from the
        input exposure's header and provide conversion functionality
        from pixels to RA/Dec and back.

        The basic syntax for using this object is::

            >>> wcs = wcsutil.WCSObject(
            ...     rootname, header=None, shape=None,
            ...     pa_key='PA_V3', new=no, prefix=None)  # doctest: +SKIP

        This will create a WCSObject which provides basic WCS functions.

        Parameters
        ==========
        rootname: string
            filename in a format supported by IRAF, specifically::

                filename.hhh[group] -or-
                filename.fits[ext] -or-
                filename.fits[extname,extver]

        header: object
            PyFITS header object from which WCS keywords can be read
        shape:    tuple
            tuple giving (nx,ny,pscale)
        pa_key: string
            name of keyword to read in telescopy orientation
        new: boolean
            specify a new object rather than creating one by
            reading in keywords from an existing image
        prefix: string
            string to use as prefix for creating archived versions
            of WCS keywords, if such keywords do not already exist

        Notes
        ======
        Setting 'new=yes' will create a WCSObject from scratch
        regardless of any input rootname.  This avoids unexpected
        filename collisions.

        Methods
        =======
        print_archive(format=True)
            print out archive keyword values
        get_archivekw(keyword)
            return archived value for WCS keyword
        set_pscale()
            set pscale attribute for object
        compute_pscale(cd11,cd21)
            compute pscale value
        get_orient()
            return orient computed from CD matrix
        updateWCS(pixel_scale=None,orient=None,refpos=None,refval=None,size=None)
            reset entire WCS based on given values
        xy2rd(pos)
            compute RA/Dec position for given (x,y) tuple
        rd2xy(skypos,hour=no)
            compute X,Y position for given (RA,Dec)
        rotateCD(orient)
            rotate CD matrix to new orientation given by 'orient'
        recenter()
            Reset reference position to X,Y center of frame
        write(fitsname=None,archive=True,overwrite=False,quiet=True)
            write out values of WCS to specified file
        restore()
            reset WCS keyword values to those from archived values
        read_archive(header,prepend=None)
            read any archive WCS keywords from PyFITS header
        archive(prepend=None,overwrite=no,quiet=yes)
            create archived copies of WCS keywords.
        write_archive(fitsname=None,overwrite=no,quiet=yes)
            write out the archived WCS values to the file
        restoreWCS(prepend=None)
            resets WCS values in file to original values
        createReferenceWCS(refname,overwrite=yes)
            write out values of WCS keywords to NEW FITS
            file without any image data
        copy(deep=True)
            create a copy of the WCSObject.
        help()
            prints out this help message

    """
    def __init__(self, rootname,header=None,shape=None,pa_key='PA_V3',new=no,prefix=None):
        # Initialize wcs dictionaries:
        #   wcsdef - default values for new images
        #   wcstrans - translation table from header keyword to attribute
        #   wcskeys  - keywords in the order they should appear in the header
        self.wcsdef = {'crpix1':0.0,'crpix2':0.0,'crval1':0.0,'crval2':0.0,'cd11':1.0,
                'cd12':1.0,'cd21':1.0,'cd22':1.0,'orient':1.0,'naxis1':0,'naxis2':0,'pscale':1.0,
                'postarg1':0.0,'postarg2':0.0,'pa_obs':0.0,
                'ctype1':'RA---TAN','ctype2':'DEC--TAN'}
        self.wcstrans = {'CRPIX1':'crpix1','CRPIX2':'crpix2','CRVAL1':'crval1','CRVAL2':'crval2',
            'CD1_1':'cd11','CD1_2':'cd12','CD2_1':'cd21','CD2_2':'cd22',
            'ORIENTAT':'orient', 'NAXIS1':'naxis1','NAXIS2':'naxis2',
            'pixel scale':'pscale','CTYPE1':'ctype1','CTYPE2':'ctype2'}
        self.wcskeys = ['NAXIS1','NAXIS2','CRPIX1','CRPIX2',
                        'CRVAL1','CRVAL2','CTYPE1','CTYPE2',
                        'CD1_1','CD1_2','CD2_1','CD2_2',
                        'ORIENTAT']
        # Now, read in the CRPIX1/2, CRVAL1/2, CD1/2_1/2 keywords.
        # Simplistic, but easy to understand what you are asking for.

        _exists = yes
        if rootname is not None:
            self.rootname = rootname
        else:
            self.rootname = 'New'
            new = yes
            _exists = no

        # Initialize attribute for GEIS image name, just in case...
        self.geisname = None

        # Look for extension specification in rootname
        _indx = _section = self.rootname.find('[')
        # If none are found, use entire rootname
        if _indx < 0:
            _indx = len(self.rootname)

        # Determine whether we are working with a new image or not.
        _dir,_rootname = os.path.split(fileutil.osfn(self.rootname[:_indx]))
        if _dir:
            _filename = _dir+os.sep+_rootname
        else:
            _filename = _rootname
        self.filename = _filename

        if not new:
            _exists = fileutil.checkFileExists(_rootname,directory=_dir)

        else:
            _exists = no

        # If no header has been provided, get the PRIMARY and the
        # specified extension header... This call uses the fully
        # expanded version of the filename, plus any sections listed by
        # by the user in the original rootname.
        if not header and _exists:
            _hdr_file = _filename+self.rootname[_indx:]
            _header = fileutil.getHeader(_hdr_file)
        else:
            # Otherwise, simply use the header already read into memory
            # for this exposure/chip.
            _header = header

        if _exists or header:
            # Initialize WCS object with keyword values...
            try:
                _dkey = 'orientat'
                if 'orientat' in _header:
                    self.orient = _header['orientat']
                else:
                    self.orient = None

                if _header['naxis'] == 0 and 'pixvalue' in _header:

                # Check for existence of NPIX/PIXVALUE keywords
                # which represent a constant array extension
                    _dkey = 'npix1'
                    self.naxis1 = _header['npix1']
                    _dkey = 'npix2'
                    self.naxis2 = _header['npix2']
                    _dkey = 'pixvalue'
                    self.pixvalue = _header['pixvalue']
                else:
                    _dkey = 'naxis1'
                    self.naxis1 = _header['naxis1']
                    _dkey = 'naxis2'
                    self.naxis2 = _header['naxis2']
                    self.pixvalue = None

                self.npix1 = self.naxis1
                self.npix2 = self.naxis2

                for key in self.wcstrans.keys():
                    _dkey = self.wcstrans[key]
                    if _dkey not in ['pscale','orient','naxis1','naxis2']:
                        self.__dict__[_dkey] = _header[key]

                self.new = no
            except:
                print('Could not find WCS keyword: ',_dkey)
                raise IOError('Image %s does not contain all required WCS keywords!' % self.rootname)

            # Now, try to read in POSTARG keyword values, if they exist...
            try:
                self.postarg1 = _header['postarg1']
                self.postarg2 = _header['postarg2']
            except:
                # If these keywords, don't exist set defaults...
                self.postarg1 = 0.0
                self.postarg2 = 0.0
            try:
                self.pa_obs = _header[pa_key]
            except:
                # If no such keyword exists, use orientat value later
                self.pa_obs = None

        else:
            # or set default values...
            self.new = yes
            for key in self.wcsdef.keys():
                self.__dict__[key] = self.wcsdef[key]

            if shape is not None:
                # ... and update with user values.
                self.naxis1 = int(shape[0])
                self.naxis2 = int(shape[1])
                self.pscale = float(shape[2])

        # Make sure reported 'orient' is consistent with CD matrix
        # while preserving the original 'ORIENTAT' keyword value
        self.orientat = self.orient

        self.orient = RADTODEG(N.arctan2(self.cd12,self.cd22))

        # If no keyword provided pa_obs value (PA_V3), then default to
        # image orientation from CD matrix.
        if self.pa_obs is None:
            self.pa_obs = self.orient

        if shape is None:
            self.set_pscale()
            #self.pscale = N.sqrt(N.power(self.cd11,2)+N.power(self.cd21,2)) * 3600.
            # Use Jacobian determination of pixel scale instead of X or Y separately...
            #self.pscale = N.sqrt(abs(self.cd11*self.cd22 - self.cd12*self.cd21))*3600.

        # Establish an attribute for the linearized orient
        # defined as the orientation of the CD after applying the default
        # distortion correction.
        self._orient_lin = 0.

        # attribute to define format for printing WCS
        self.__format__=yes

        # Keep track of the keyword names used as the backup keywords
        # for the original WCS values
        #    backup - dict relating active keywords with backup keywords
        #    prepend - string prepended to active keywords to create backup keywords
        #    orig_wcs - dict containing orig keywords and values
        self.backup = {}
        self.revert = {}
        self.prepend = None
        self.orig_wcs = {}
        # Read in any archived WCS keyword values, if they exist
        self.read_archive(_header,prepend=prefix)

    # You never know when you want to print out the WCS keywords...
    def __str__(self):
        block = 'WCS Keywords for ' + self.rootname + ': \n'
        if not self.__format__:
            for key in self.wcstrans.keys():
                _dkey = self.wcstrans[key]
                strn = key.upper() + " = " + repr(self.__dict__[_dkey]) + '\n'
                block += strn
            block += 'PA_V3: '+repr(self.pa_obs)+'\n'

        else:
            block += 'CD_11  CD_12: '+repr(self.cd11)+'  '+repr(self.cd12) +'\n'
            block += 'CD_21  CD_22: '+repr(self.cd21)+'  '+repr(self.cd22) +'\n'
            block += 'CRVAL       : '+repr(self.crval1)+'  '+repr(self.crval2) + '\n'
            block += 'CRPIX       : '+repr(self.crpix1)+'  '+repr(self.crpix2) + '\n'
            block += 'NAXIS       : '+repr(int(self.naxis1))+'  '+repr(int(self.naxis2)) + '\n'
            block += 'Plate Scale : '+repr(self.pscale)+'\n'
            block += 'ORIENTAT    : '+repr(self.orient)+'\n'
            block += 'CTYPE       : '+repr(self.ctype1)+'  '+repr(self.ctype2)+'\n'
            block += 'PA Telescope: '+repr(self.pa_obs)+'\n'

        return block

    def __repr__(self):
        return repr(self.__dict__)

    def print_archive(self,format=True):
        """ Prints out archived WCS keywords."""
        if len(list(self.orig_wcs.keys())) > 0:
            block  = 'Original WCS keywords for ' + self.rootname+ '\n'
            block += '    backed up on '+repr(self.orig_wcs['WCSCDATE'])+'\n'
            if not format:
                for key in self.wcstrans.keys():
                    block += key.upper() + " = " + repr(self.get_archivekw(key)) + '\n'
                block = 'PA_V3: '+repr(self.pa_obs)+'\n'

            else:
                block += 'CD_11  CD_12: '+repr(self.get_archivekw('CD1_1'))+'  '+repr(self.get_archivekw('CD1_2')) +'\n'
                block += 'CD_21  CD_22: '+repr(self.get_archivekw('CD2_1'))+'  '+repr(self.get_archivekw('CD2_2')) +'\n'
                block += 'CRVAL       : '+repr(self.get_archivekw('CRVAL1'))+'  '+repr(self.get_archivekw('CRVAL2')) + '\n'
                block += 'CRPIX       : '+repr(self.get_archivekw('CRPIX1'))+'  '+repr(self.get_archivekw('CRPIX2')) + '\n'
                block += 'NAXIS       : '+repr(int(self.get_archivekw('NAXIS1')))+'  '+repr(int(self.get_archivekw('NAXIS2'))) + '\n'
                block += 'Plate Scale : '+repr(self.get_archivekw('pixel scale'))+'\n'
                block += 'ORIENTAT    : '+repr(self.get_archivekw('ORIENTAT'))+'\n'

            print(block)

    def get_archivekw(self,keyword):
        """ Return an archived/backup value for the keyword. """
        return self.orig_wcs[self.backup[keyword]]

    def set_pscale(self):
        """ Compute the pixel scale based on active WCS values. """
        if self.new:
            self.pscale = 1.0
        else:
            self.pscale = self.compute_pscale(self.cd11,self.cd21)

    def compute_pscale(self,cd11,cd21):
        """ Compute the pixel scale based on active WCS values. """
        return N.sqrt(N.power(cd11,2)+N.power(cd21,2)) * 3600.

    def get_orient(self):
        """ Return the computed orientation based on CD matrix. """
        return RADTODEG(N.arctan2(self.cd12,self.cd22))

    def set_orient(self):
        """ Return the computed orientation based on CD matrix. """
        self.orient = RADTODEG(N.arctan2(self.cd12,self.cd22))

    def update(self):
        """ Update computed values of WCS based on current CD matrix."""
        self.set_pscale()
        self.set_orient()

    def updateWCS(self, pixel_scale=None, orient=None,refpos=None,refval=None,size=None):
        """
        Create a new CD Matrix from the absolute pixel scale
        and reference image orientation.
        """
        # Set up parameters necessary for updating WCS
        # Check to see if new value is provided,
        # If not, fall back on old value as the default

        _updateCD = no
        if orient is not None and orient != self.orient:
            pa = DEGTORAD(orient)
            self.orient = orient
            self._orient_lin = orient
            _updateCD = yes
        else:
            # In case only pixel_scale was specified
            pa = DEGTORAD(self.orient)

        if pixel_scale is not None and pixel_scale != self.pscale:
            _ratio = pixel_scale / self.pscale
            self.pscale = pixel_scale
            _updateCD = yes
        else:
            # In case, only orient was specified
            pixel_scale = self.pscale
            _ratio = None

        # If a new plate scale was given,
        # the default size should be revised accordingly
        # along with the default reference pixel position.
        # Added 31 Mar 03, WJH.
        if _ratio is not None:
            self.naxis1 /= _ratio
            self.naxis2 /= _ratio
            self.crpix1 = self.naxis1/2.
            self.crpix2 = self.naxis2/2.

        # However, if the user provides a given size,
        # set it to use that no matter what.
        if size is not None:
            self.naxis1 = size[0]
            self.naxis2 = size[1]

        # Insure that naxis1,2 always return as integer values.
        self.naxis1 = int(self.naxis1)
        self.naxis2 = int(self.naxis2)

        if refpos is not None:
            self.crpix1 = refpos[0]
            self.crpix2 = refpos[1]
        if self.crpix1 is None:
            self.crpix1 = self.naxis1/2.
            self.crpix2 = self.naxis2/2.

        if refval is not None:
            self.crval1 = refval[0]
            self.crval2 = refval[1]

        # Reset WCS info now...
        if _updateCD:
            # Only update this should the pscale or orientation change...
            pscale = pixel_scale / 3600.

            self.cd11 = -pscale * N.cos(pa)
            self.cd12 = pscale * N.sin(pa)
            self.cd21 = self.cd12
            self.cd22 = -self.cd11

        # Now make sure that all derived values are really up-to-date based
        # on these changes
        self.update()

    def scale_WCS(self,pixel_scale,retain=True):
        ''' Scale the WCS to a new pixel_scale. The 'retain' parameter
        [default value: True] controls whether or not to retain the original
        distortion solution in the CD matrix.
        '''
        _ratio = pixel_scale / self.pscale

        # Correct the size of the image and CRPIX values for scaled WCS
        self.naxis1 /= _ratio
        self.naxis2 /= _ratio
        self.crpix1 = self.naxis1/2.
        self.crpix2 = self.naxis2/2.

        if retain:
            # Correct the WCS while retaining original distortion information
            self.cd11 *= _ratio
            self.cd12 *= _ratio
            self.cd21 *= _ratio
            self.cd22 *= _ratio
        else:
            pscale = pixel_scale / 3600.
            self.cd11 = -pscale * N.cos(pa)
            self.cd12 = pscale * N.sin(pa)
            self.cd21 = self.cd12
            self.cd22 = -self.cd11

        # Now make sure that all derived values are really up-to-date based
        # on these changes
        self.update()

    def xy2rd(self,pos):
        """
        This method would apply the WCS keywords to a position to
        generate a new sky position.

        The algorithm comes directly from 'imgtools.xy2rd'

        translate (x,y) to (ra, dec)
        """
        if self.ctype1.find('TAN') < 0 or self.ctype2.find('TAN') < 0:
            print('XY2RD only supported for TAN projections.')
            raise TypeError

        if isinstance(pos,N.ndarray):
            # If we are working with an array of positions,
            # point to just X and Y values
            posx = pos[:,0]
            posy = pos[:,1]
        else:
            # Otherwise, we are working with a single X,Y tuple
            posx = pos[0]
            posy = pos[1]

        xi = self.cd11 * (posx - self.crpix1) + self.cd12 * (posy - self.crpix2)
        eta = self.cd21 * (posx - self.crpix1) + self.cd22 * (posy - self.crpix2)

        xi = DEGTORAD(xi)
        eta = DEGTORAD(eta)
        ra0 = DEGTORAD(self.crval1)
        dec0 = DEGTORAD(self.crval2)

        ra = N.arctan((xi / (N.cos(dec0)-eta*N.sin(dec0)))) + ra0
        dec = N.arctan( ((eta*N.cos(dec0)+N.sin(dec0)) /
                (N.sqrt((N.cos(dec0)-eta*N.sin(dec0))**2 + xi**2))) )

        ra = RADTODEG(ra)
        dec = RADTODEG(dec)
        ra = DIVMOD(ra, 360.)

        # Otherwise, just return the RA,Dec tuple.
        return ra,dec


    def rd2xy(self,skypos,hour=no):
        """
        This method would use the WCS keywords to compute the XY position
        from a given RA/Dec tuple (in deg).

        NOTE: Investigate how to let this function accept arrays as well
        as single positions. WJH 27Mar03

        """
        if self.ctype1.find('TAN') < 0 or self.ctype2.find('TAN') < 0:
            print('RD2XY only supported for TAN projections.')
            raise TypeError

        det = self.cd11*self.cd22 - self.cd12*self.cd21

        if det == 0.0:
            raise ArithmeticError("singular CD matrix!")

        cdinv11 = self.cd22 / det
        cdinv12 = -self.cd12 / det
        cdinv21 = -self.cd21 / det
        cdinv22 = self.cd11 / det

        # translate (ra, dec) to (x, y)

        ra0 = DEGTORAD(self.crval1)
        dec0 = DEGTORAD(self.crval2)
        if hour:
            skypos[0] = skypos[0] * 15.
        ra = DEGTORAD(skypos[0])
        dec = DEGTORAD(skypos[1])

        bottom = float(N.sin(dec)*N.sin(dec0) + N.cos(dec)*N.cos(dec0)*N.cos(ra-ra0))
        if bottom == 0.0:
            raise ArithmeticError("Unreasonable RA/Dec range!")

        xi = RADTODEG((N.cos(dec) * N.sin(ra-ra0) / bottom))
        eta = RADTODEG((N.sin(dec)*N.cos(dec0) - N.cos(dec)*N.sin(dec0)*N.cos(ra-ra0)) / bottom)

        x = cdinv11 * xi + cdinv12 * eta + self.crpix1
        y = cdinv21 * xi + cdinv22 * eta + self.crpix2

        return x,y

    def rotateCD(self,orient):
        """ Rotates WCS CD matrix to new orientation given by 'orient'
        """
        # Determine where member CRVAL position falls in ref frame
        # Find out whether this needs to be rotated to align with
        # reference frame.

        _delta = self.get_orient() - orient
        if _delta == 0.:
            return

        # Start by building the rotation matrix...
        _rot = fileutil.buildRotMatrix(_delta)
        # ...then, rotate the CD matrix and update the values...
        _cd = N.array([[self.cd11,self.cd12],[self.cd21,self.cd22]],dtype=N.float64)
        _cdrot = N.dot(_cd,_rot)
        self.cd11 = _cdrot[0][0]
        self.cd12 = _cdrot[0][1]
        self.cd21 = _cdrot[1][0]
        self.cd22 = _cdrot[1][1]
        self.orient = orient

    def recenter(self):
        """
        Reset the reference position values to correspond to the center
        of the reference frame.
        Algorithm used here developed by Colin Cox - 27-Jan-2004.
        """
        if self.ctype1.find('TAN') < 0 or self.ctype2.find('TAN') < 0:
            print('WCS.recenter() only supported for TAN projections.')
            raise TypeError

        # Check to see if WCS is already centered...
        if self.crpix1 == self.naxis1/2. and self.crpix2 == self.naxis2/2.:
            # No recentering necessary... return without changing WCS.
            return

        # This offset aligns the WCS to the center of the pixel, in accordance
        # with the 'align=center' option used by 'drizzle'.
        #_drz_off = -0.5
        _drz_off = 0.
        _cen = (self.naxis1/2.+ _drz_off,self.naxis2/2. + _drz_off)

        # Compute the RA and Dec for center pixel
        _cenrd = self.xy2rd(_cen)
        _cd = N.array([[self.cd11,self.cd12],[self.cd21,self.cd22]],dtype=N.float64)
        _ra0 = DEGTORAD(self.crval1)
        _dec0 = DEGTORAD(self.crval2)
        _ra = DEGTORAD(_cenrd[0])
        _dec = DEGTORAD(_cenrd[1])

        # Set up some terms for use in the final result
        _dx = self.naxis1/2. - self.crpix1
        _dy = self.naxis2/2. - self.crpix2

        _dE,_dN = DEGTORAD(N.dot(_cd,(_dx,_dy)))
        _dE_dN = 1 + N.power(_dE,2) + N.power(_dN,2)
        _cosdec = N.cos(_dec)
        _sindec = N.sin(_dec)
        _cosdec0 = N.cos(_dec0)
        _sindec0 = N.sin(_dec0)

        _n1 = N.power(_cosdec,2) + _dE*_dE + _dN*_dN*N.power(_sindec,2)
        _dra_dE = (_cosdec0 - _dN*_sindec0)/_n1
        _dra_dN = _dE*_sindec0 /_n1

        _ddec_dE = -_dE*N.tan(_dec) / _dE_dN
        _ddec_dN = (1/_cosdec) * ((_cosdec0 / N.sqrt(_dE_dN)) - (_dN*N.sin(_dec) / _dE_dN))

        # Compute new CD matrix values now...
        _cd11n = _cosdec * (self.cd11*_dra_dE + self.cd21 * _dra_dN)
        _cd12n = _cosdec * (self.cd12*_dra_dE + self.cd22 * _dra_dN)
        _cd21n = self.cd11 * _ddec_dE + self.cd21 * _ddec_dN
        _cd22n = self.cd12 * _ddec_dE + self.cd22 * _ddec_dN

        _new_orient = RADTODEG(N.arctan2(_cd12n,_cd22n))
        #_new_pscale = N.sqrt(N.power(_cd11n,2)+N.power(_cd21n,2)) * 3600.

        # Update the values now...
        self.crpix1 = _cen[0]
        self.crpix2 = _cen[1]
        self.crval1 = RADTODEG(_ra)
        self.crval2 = RADTODEG(_dec)

        # Keep the same plate scale, only change the orientation
        self.rotateCD(_new_orient)

        # These would update the CD matrix with the new rotation
        # ALONG with the new plate scale which we do not want.
        self.cd11 = _cd11n
        self.cd12 = _cd12n
        self.cd21 = _cd21n
        self.cd22 = _cd22n
        #self.pscale = _new_pscale

        self.update()

    def write(self,fitsname=None,wcs=None,archive=True,overwrite=False,quiet=True):
        """
        Write out the values of the WCS keywords to the
        specified image.

        If it is a GEIS image and 'fitsname' has been provided,
        it will automatically make a multi-extension
        FITS copy of the GEIS and update that file. Otherwise, it
        throw an Exception if the user attempts to directly update
        a GEIS image header.

        If archive=True, also write out archived WCS keyword values to file.
        If overwrite=True, replace archived WCS values in file with new values.

        If a WCSObject is passed through the 'wcs' keyword, then the WCS keywords
        of this object are copied to the header of the image to be updated. A use case
        fo rthis is updating the WCS of a WFPC2 data quality (_c1h.fits) file
        in order to be in sync with the science (_c0h.fits) file.

        """
        ## Start by making sure all derived values are in sync with CD matrix
        self.update()

        image = self.rootname
        _fitsname = fitsname

        if image.find('.fits') < 0 and _fitsname is not None:
            # A non-FITS image was provided, and openImage made a copy
            # Update attributes to point to new copy instead
            self.geisname = image
            image = self.rootname = _fitsname

        # Open image as writable FITS object
        fimg = fileutil.openImage(image, mode='update', fitsname=_fitsname)

        _root,_iextn = fileutil.parseFilename(image)
        _extn = fileutil.getExtn(fimg,_iextn)

        # Write out values to header...
        if wcs:
            _wcsobj = wcs
        else:
            _wcsobj = self

        for key in _wcsobj.wcstrans.keys():
            _dkey = _wcsobj.wcstrans[key]
            if _dkey != 'pscale':
                _extn.header[key] = _wcsobj.__dict__[_dkey]

        # Close the file
        fimg.close()
        del fimg
        if archive:
            self.write_archive(fitsname=fitsname,overwrite=overwrite,quiet=quiet)

    def restore(self):
        """ Reset the active WCS keywords to values stored in the
            backup keywords.
        """
        # If there are no backup keys, do nothing...
        if len(list(self.backup.keys())) == 0:
            return
        for key in self.backup.keys():
            if key != 'WCSCDATE':
                self.__dict__[self.wcstrans[key]] = self.orig_wcs[self.backup[key]]

        self.update()

    def archive(self,prepend=None,overwrite=no,quiet=yes):
        """ Create backup copies of the WCS keywords with the given prepended
            string.
            If backup keywords are already present, only update them if
            'overwrite' is set to 'yes', otherwise, do warn the user and do nothing.
            Set the WCSDATE at this time as well.
        """
        # Verify that existing backup values are not overwritten accidentally.
        if len(list(self.backup.keys())) > 0 and overwrite == no:
            if not quiet:
                print('WARNING: Backup WCS keywords already exist! No backup made.')
                print('         The values can only be overridden if overwrite=yes.')
            return

        # Establish what prepend string to use...
        if prepend is None:
            if self.prepend is not None:
                _prefix = self.prepend
            else:
                _prefix = DEFAULT_PREFIX
        else:
            _prefix = prepend

        # Update backup and orig_wcs dictionaries
        # We have archive keywords and a defined prefix
        # Go through and append them to self.backup
        self.prepend = _prefix
        for key in self.wcstrans.keys():
            if key != 'pixel scale':
                _archive_key = self._buildNewKeyname(key,_prefix)
            else:
                _archive_key = self.prepend.lower()+'pscale'
#            if key != 'pixel scale':
            self.orig_wcs[_archive_key] = self.__dict__[self.wcstrans[key]]
            self.backup[key] = _archive_key
            self.revert[_archive_key] = key

        # Setup keyword to record when these keywords were backed up.
        self.orig_wcs['WCSCDATE']= fileutil.getLTime()
        self.backup['WCSCDATE'] = 'WCSCDATE'
        self.revert['WCSCDATE'] = 'WCSCDATE'

    def read_archive(self,header,prepend=None):
        """ Extract a copy of WCS keywords from an open file header,
            if they have already been created and remember the prefix
            used for those keywords. Otherwise, setup the current WCS
            keywords as the archive values.
        """
        # Start by looking for the any backup WCS keywords to
        # determine whether archived values are present and to set
        # the prefix used.
        _prefix = None
        _archive = False
        if header is not None:
            for kw in header.items():
                if kw[0][1:] in self.wcstrans.keys():
                    _prefix = kw[0][0]
                    _archive = True
                    break

        if not _archive:
            self.archive(prepend=prepend)
            return

        # We have archive keywords and a defined prefix
        # Go through and append them to self.backup
        if _prefix is not None:
            self.prepend = _prefix
        else:
            self.prepend = DEFAULT_PREFIX

        for key in self.wcstrans.keys():
            _archive_key = self._buildNewKeyname(key,_prefix)
            if key!= 'pixel scale':
                if _archive_key in header:
                    self.orig_wcs[_archive_key] = header[_archive_key]
                else:
                    self.orig_wcs[_archive_key] = header[key]
                self.backup[key] = _archive_key
                self.revert[_archive_key] = key

        # Establish plate scale value
        _cd11str = self.prepend+'CD1_1'
        _cd21str = self.prepend+'CD2_1'
        pscale = self.compute_pscale(self.orig_wcs[_cd11str],self.orig_wcs[_cd21str])
        _archive_key = self.prepend.lower()+'pscale'
        self.orig_wcs[_archive_key] = pscale
        self.backup['pixel scale'] = _archive_key
        self.revert[_archive_key] = 'pixel scale'

        # Setup keyword to record when these keywords were backed up.
        if 'WCSCDATE' in header:
            self.orig_wcs['WCSCDATE'] = header['WCSCDATE']
        else:
            self.orig_wcs['WCSCDATE'] = fileutil.getLTime()
        self.backup['WCSCDATE'] = 'WCSCDATE'
        self.revert['WCSCDATE'] = 'WCSCDATE'

    def write_archive(self,fitsname=None,overwrite=no,quiet=yes):
        """ Saves a copy of the WCS keywords from the image header
            as new keywords with the user-supplied 'prepend'
            character(s) prepended to the old keyword names.

            If the file is a GEIS image and 'fitsname' is not None, create
            a FITS copy and update that version; otherwise, raise
            an Exception and do not update anything.

        """
        _fitsname = fitsname

        # Open image in update mode
        #    Copying of GEIS images handled by 'openImage'.
        fimg = fileutil.openImage(self.rootname,mode='update',fitsname=_fitsname)
        if self.rootname.find('.fits') < 0 and _fitsname is not None:
            # A non-FITS image was provided, and openImage made a copy
            # Update attributes to point to new copy instead
            self.geisname = self.rootname
            self.rootname = _fitsname

        # extract the extension ID being updated
        _root,_iextn = fileutil.parseFilename(self.rootname)
        _extn = fileutil.getExtn(fimg,_iextn)
        if not quiet:
            print('Updating archive WCS keywords for ',_fitsname)

        # Write out values to header...
        for key in self.orig_wcs.keys():
            _comment = None
            _dkey = self.revert[key]

            # Verify that archive keywords will not be overwritten,
            # unless overwrite=yes.
            _old_key = key in _extn.header
            if  _old_key == True and overwrite == no:
                if not quiet:
                    print('WCS keyword',key,' already exists! Not overwriting.')
                continue

            # No archive keywords exist yet in file, or overwrite=yes...
            # Extract the value for the original keyword
            if _dkey in _extn.header:

                # Extract any comment string for the keyword as well
                _indx_key = _extn.header.index(_dkey)
                _full_key = _extn.header.cards[_indx_key]
                if not quiet:
                    print('updating ',key,' with value of: ',self.orig_wcs[key])
                _extn.header[key] = (self.orig_wcs[key], _full_key.comment)

        key = 'WCSCDATE'
        if key not in _extn.header:
            # Print out history keywords to record when these keywords
            # were backed up.
            _extn.header[key] = (self.orig_wcs[key], "Time WCS keywords were copied.")

        # Close the now updated image
        fimg.close()
        del fimg

    def restoreWCS(self,prepend=None):
        """ Resets the WCS values to the original values stored in
            the backup keywords recorded in self.backup.
        """
        # Open header for image
        image = self.rootname

        if prepend: _prepend = prepend
        elif self.prepend: _prepend = self.prepend
        else: _prepend = None

        # Open image as writable FITS object
        fimg = fileutil.openImage(image, mode='update')
        # extract the extension ID being updated
        _root,_iextn = fileutil.parseFilename(self.rootname)
        _extn = fileutil.getExtn(fimg,_iextn)

        if len(self.backup) > 0:
            # If it knows about the backup keywords already,
            # use this to restore the original values to the original keywords
            for newkey in self.revert.keys():
                if newkey != 'opscale':
                    _orig_key = self.revert[newkey]
                    _extn.header[_orig_key] = _extn.header[newkey]
        elif _prepend:
            for key in self.wcstrans.keys():
                # Get new keyword name based on old keyname
                #    and prepend string
                if key != 'pixel scale':
                    _okey = self._buildNewKeyname(key,_prepend)

                    if _okey in _extn.header:
                        _extn.header[key] = _extn.header[_okey]
                    else:
                        print('No original WCS values found. Exiting...')
                        break
        else:
            print('No original WCS values found. Exiting...')

        fimg.close()
        del fimg

    def createReferenceWCS(self,refname,overwrite=yes):
        """ Write out the values of the WCS keywords to the NEW
            specified image 'fitsname'.

        """
        hdu = self.createWcsHDU()
        # If refname already exists, delete it to make way for new file
        if os.path.exists(refname):
            if overwrite==yes:
                # Remove previous version and re-create with new header
                os.remove(refname)
                hdu.writeto(refname)
            else:
                # Append header to existing file
                wcs_append = True
                oldhdu = fits.open(refname, mode='append')
                for e in oldhdu:
                    if 'extname' in e.header and e.header['extname'] == 'WCS':
                        wcs_append = False
                if wcs_append == True:
                    oldhdu.append(hdu)
                oldhdu.close()
                del oldhdu
        else:
            # No previous file, so generate new one from scratch
            hdu.writeto(refname)

        # Clean up
        del hdu

    def createWcsHDU(self):
        """ Generate a WCS header object that can be used to
            populate a reference WCS HDU.
        """
        hdu = fits.ImageHDU()
        hdu.header['EXTNAME'] = 'WCS'
        hdu.header['EXTVER'] = 1
        # Now, update original image size information
        hdu.header['WCSAXES'] = (2, "number of World Coordinate System axes")
        hdu.header['NPIX1'] = (self.naxis1, "Length of array axis 1")
        hdu.header['NPIX2'] = (self.naxis2, "Length of array axis 2")
        hdu.header['PIXVALUE'] = (0.0, "values of pixels in array")

        # Write out values to header...
        excluded_keys = ['naxis1','naxis2']
        for key in self.wcskeys:
            _dkey = self.wcstrans[key]
            if _dkey not in excluded_keys:
                hdu.header[key] = self.__dict__[_dkey]


        return hdu

    def _buildNewKeyname(self,key,prepend):
        """ Builds a new keyword based on original keyword name and
            a prepend string.
        """

        if len(prepend+key) <= 8: _new_key = prepend+key
        else: _new_key = str(prepend+key)[:8]

        return _new_key


    def copy(self,deep=yes):
        """ Makes a (deep)copy of this object for use by other objects.
        """
        if deep:
            return copy.deepcopy(self)
        else:
            return copy.copy(self)

    def help(self):
        """ Prints out help message."""
        print('wcsutil Version '+str(__version__)+':\n')
        print(self.__doc__)



"""module 'editpar.py' -- main module for generating the EPAR task editor

$Id$

Taken from pyraf/lib/epar.py, originally signed "M.D. De La Pena, 2000 Feb. 4"
"""
from __future__ import absolute_import, division, print_function # confidence high

#System level modules
import os, sys, tempfile, time
from . import capable

PY3K = sys.version_info[0] > 2

if capable.OF_GRAPHICS:
    if PY3K:
        from tkinter import  _default_root
        from tkinter import *
        from tkinter.filedialog import asksaveasfilename
        from tkinter.messagebox import askokcancel, askyesno, showwarning
    else:
        from Tkinter import  _default_root
        from Tkinter import *
        from tkFileDialog import asksaveasfilename
        from tkMessageBox import askokcancel, askyesno, showwarning

# stsci.tools modules
from .irafglobals import userWorkingHome
from . import basicpar, eparoption, irafutils, taskpars

# Constants
MINVIEW     = 500
MINPARAMS   = 25
INPUTWIDTH  = 10
VALUEWIDTH  = 21
PROMPTWIDTH = 55
DFT_OPT_FILE = "epar.optionDB"
TIP = "rollover"
DBG = "debug"

# The following action types are used within the GUI code.  They define what
# kind of GUI action actually caused a parameter's value to be adjusted.
# This is meant to be like an enum.  These values may appear in a task's
# task.cfgspc file in a rule.  In that file, the value 'always' may be used, in
# addition to these values, to indicate a match to all possible action types.
GROUP_ACTIONS = ('defaults','init','fopen','entry')
# init     -> startup of the GUI
# defaults -> the user clicked the Defaults or Reset button
# fopen    -> the user loaded a config file
# entry    -> the user actually edited a parameter (via mouse or keyboard)

# Use these values for startup geometry ***for now***
# PARENT is the main editor window
PARENTX = 50
PARENTY = 50

# DCHILD[XY] are amounts each successive child shifts
DCHILDX = 50
DCHILDY = 50

# CHILD[XY] is a PSET window
CHILDX = PARENTX
CHILDY = PARENTY

# HELP[XY] is for the help as displayed in a window
HELPX   = 300
HELPY   = 25


class UnfoundParamError(Exception): pass


class EditParDialog(object):

    def __init__(self, theTask, parent=None, isChild=0,
                 title="Parameter Editor", childList=None,
                 resourceDir='.'):

        # Initialize status message stuff first thing
        self._leaveStatusMsgUntil = 0
        self._msgHistory = [] # all msgs, of all kinds, since we opened
        self._statusMsgsToShow = [] # keep a *small* number of late msgs
        self.debug('Starting up the GUI!')

        # Call our (or a subclass's) _setTaskParsObj() method
        self._setTaskParsObj(theTask)

        # Now go back and ensure we have the full taskname; set up other items
        self._canceled = False
        self._executed = False
        self._guiName = title
        self.taskName = self._taskParsObj.getName()
        self.pkgName = self._taskParsObj.getPkgname()
        theParamList = self._taskParsObj.getParList(docopy=1)
        self._rcDir = resourceDir
        self.debug('TASK: '+self.taskName+', PKG: '+self.pkgName+ \
                   ', RC: '+self._rcDir)
        # setting _tmwm=1 is the slowest motion, 7 seems OK, 10 maybe too fast
        self._tmwm = int(os.getenv('TEAL_MOUSE_WHEEL_MULTIPLIER', 7))

        # Get default parameter values for unlearn - watch return value
        # NOTE - this may edit/reorder the working paramList
        if not self._setupDefaultParamList():
            return

        # Ignore the last parameter which is $nargs
        self.numParams = len(theParamList) - 1

        # Set all default master GUI settings, then
        # allow subclasses to override them
        self._appName             = "Par Editor"
        self._appHelpString       = "No help yet created for this GUI editor"
        self._useSimpleAutoClose  = False # certain buttons close GUI also
        self._showExecuteButton   = True
        self._showSaveCloseOnExec = True
        self._saveAndCloseOnExec  = True
        self._showFlaggingChoice  = True
        self._flagNonDefaultVals  = None # default not yet set
        self._showExtraHelpButton = False
        self._showHelpInBrowser   = False
        self._knowTaskHelpIsHtml  = False
        self._unpackagedTaskTitle = "Task"
        self._writeProtectOnSaveAs= True
        self._defaultsButtonTitle = "Defaults"
        self._optFile             = DFT_OPT_FILE
        self._defSaveAsExt        = '.cfg'

        # Colors
        self._frmeColor = None  # frame of window
        self._taskColor = None  # task label area
        self._bboxColor = None  # button area
        self._entsColor = None  # entries area
        self._flagColor = "red" # non-default values

        # give the subclass a chance to disagree
        self._overrideMasterSettings() # give the subclass a chance to disagree

        # any settings which depend on overrides
        if self._flagNonDefaultVals is None:
            self._flagNonDefaultVals = self._showFlaggingChoice # default

        # Create the root window as required, but hide it
        self.parent = parent
        if self.parent is None:
            global _default_root
            if _default_root is None:
                _default_root = irafutils.init_tk_default_root()

        # Track whether this is a parent or child window
        self.isChild = isChild

        # Set up a color for each of the backgrounds
        if self.isChild:
        #    self._frmeColor = "LightSteelBlue"
            self.iconLabel = "EPAR Child"
        else:
            self.iconLabel = "EPAR Parent"

        # help windows do not exist yet
        self.eparHelpWin = None
        self.irafHelpWin = None
        self.logHistWin = None

        # no last focus widget
        self.lastFocusWidget = None

        # Generate the top epar window
        self.top = top = Toplevel(self.parent,bg=self._frmeColor,visual="best")
        self.top.withdraw() # hide it while we fill it up with stuff

        if len(self.pkgName):
            self.updateTitle(self.pkgName+"."+self.taskName)
        else:
            self.updateTitle(self.taskName)
        self.top.iconname(self.iconLabel)

        # Read in the tk options database file
        try:
            # User's current directory
            self.top.option_readfile(os.path.join(os.curdir, self._optFile))
        except TclError:
            try:
                # User's startup directory
                self.top.option_readfile(os.path.join(userWorkingHome,
                                                      self._optFile))
            except TclError:
                try:
                    # App default
                    self.top.option_readfile(os.path.join(self._rcDir,
                                                          self._optFile))
                except TclError:
                    if self._optFile != DFT_OPT_FILE:
                        pass
                    else:
                        raise

        # Create an empty list to hold child dialogs
        # *** Not a good way, REDESIGN with Mediator!
        # Also, build the parent menu bar
        if self.parent is None:
            self.top.childList = []
        elif childList is not None:
            # all children share a list
            self.top.childList = childList

        # Build the EPAR menu bar
        self.makeMenuBar(self.top)

        # Create a spacer
        Frame(self.top, bg=self._taskColor, height=10).pack(side=TOP, fill=X)

        # Print the package and task names
        self.printNames(self.top, self.taskName, self.pkgName)

        # Insert a spacer between the static text and the buttons
        Frame(self.top, bg=self._taskColor, height=15).pack(side=TOP, fill=X)

        # Set control buttons at the top of the frame
        self.buttonBox(self.top)

        # Insert a spacer between the static text and the buttons
        Frame(self.top, bg=self._entsColor, height=15).pack(side=TOP, fill=X)

        # Set up an information Frame at the bottom of the EPAR window
        # RESIZING is currently disabled.
        # Do this here so when resizing to a smaller sizes, the parameter
        # panel is reduced - not the information frame.
        self.top.status = Label(self.top, text="", relief=SUNKEN,
                                borderwidth=1, anchor=W, bg=self._frmeColor)
        self.top.status.pack(side=BOTTOM, fill=X, padx=0, pady=3, ipady=3)

        # Set up a Frame to hold a scrollable Canvas
        self.top.f = frame = Frame(self.top, relief=RIDGE, borderwidth=1,
                                   bg=self._entsColor)

        # Overlay a Canvas which will hold a Frame
        self.top.f.canvas = canvas = Canvas(self.top.f, width=100, height=100,
            takefocus=FALSE, bg=self._entsColor,
            highlightbackground=self._entsColor)
#           highlightcolor="black" # black must be the default, since it is blk

        # Always build the scrollbar, even if number of parameters is small,
        # to allow window to be resized.

        # Attach a vertical Scrollbar to the Frame/Canvas
        self.top.f.vscroll = Scrollbar(self.top.f, orient=VERTICAL,
             width=11, relief=SUNKEN, activerelief=RAISED,
             takefocus=FALSE, bg=self._entsColor)
        canvas['yscrollcommand'] = self.top.f.vscroll.set
        self.top.f.vscroll['command'] = canvas.yview

        # Pack the Scrollbar
        self.top.f.vscroll.pack(side=RIGHT, fill=Y)

        # enable Page Up/Down keys
        scroll = canvas.yview_scroll
        top.bind('<Next>', lambda event, fs=scroll: fs(1, "pages"))
        top.bind('<Prior>', lambda event, fs=scroll: fs(-1, "pages"))

        # make up, down arrows and return/shift-return do same as Tab, Shift-Tab
        top.bind('<Up>', self.focusPrev)
        top.bind('<MouseWheel>', self.mwl) # on OSX, rolled up or down
        top.bind('<Button-4>', self.mwl)   # on Linux, rolled up
        top.bind('<Button-5>', self.mwl)   # on Linux, rolled down
        top.bind('<Down>', self.focusNext)
        top.bind('<Shift-Return>', self.focusPrev)
        top.bind('<Return>', self.focusNext)
        try:
            # special shift-tab binding needed for (some? all?) linux systems
            top.bind('<KeyPress-ISO_Left_Tab>', self.focusPrev)
        except TclError:
            # Ignore exception here, the binding can't be relevant
            # if ISO_Left_Tab is unknown.
            pass

        # Pack the Frame and Canvas
        canvas.pack(side=TOP, expand=TRUE, fill=BOTH)
        self.top.f.pack(side=TOP, expand=TRUE, fill=BOTH)

        # Define a Frame to contain the parameter information
        canvas.entries = Frame(canvas, bg=self._entsColor)

        # Generate the window to hold the Frame which sits on the Canvas
        cWindow = canvas.create_window(0, 0,
                           anchor=NW,
                           window=canvas.entries)

        # Insert a spacer between the Canvas and the information frame
        Frame(self.top, bg=self._entsColor, height=4).pack(side=TOP, fill=X)

        # The parent has the control, unless there are children
        # Fix the geometry of where the windows first appear on the screen
        if self.parent is None:
            #self.top.grab_set()

            # Position this dialog relative to the parent
            self.top.geometry("+%d+%d" % (PARENTX, PARENTY))
        else:
            #self.parent.grab_release()
            #self.top.grab_set()

            # Declare the global variables so they can be updated
            global CHILDX
            global CHILDY

            # Position this dialog relative to the parent
            CHILDX = CHILDX + DCHILDX
            CHILDY = CHILDY + DCHILDY
            self.top.geometry("+%d+%d" % (CHILDX, CHILDY))


        #
        # Now fill in the Canvas Window
        #

        # The makeEntries method creates the parameter entry Frame
        self.makeEntries(canvas.entries, self.top.status)

        # Force an update of the entry Frame
        canvas.entries.update()

        # Determine the size of the entry Frame
        width = canvas.entries.winfo_width()
        height = canvas.entries.winfo_height()

        # Reconfigure the Canvas size based on the Frame.
        if (self.numParams <= MINPARAMS):
            viewHeight = height
        else:
            # Set the minimum display
            viewHeight = MINVIEW

        # Scrollregion is based upon the full size of the entry Frame
        canvas.config(scrollregion=(0, 0, width, height))
        # Smooth scroll
        self.yscrollincrement = 5 # changed Mar2010, had been 50 a long time
        canvas.config(yscrollincrement=self.yscrollincrement)

        # Set the actual viewable region for the Canvas
        canvas.config(width=width, height=viewHeight)

        # Force an update of the Canvas
        canvas.update()

        # Associate deletion of the main window to a Abort
        self.top.protocol("WM_DELETE_WINDOW", self.abort)

        # Trigger all widgets one time before starting in case they have
        # values which would run a trigger
        self.checkAllTriggers('init')

        # Set focus to first parameter
        self.setViewAtTop()

        # Finally show it
        self.top.update()
        self.top.deiconify()

        # Enable interactive resizing in height
        self.top.resizable(width=FALSE, height=TRUE)

        # Limit maximum window height
        width = self.top.winfo_width()
        height = self.top.winfo_height() + height - viewHeight
        self.top.maxsize(width=width, height=height)

        self.debug('showing '+self._appName+' main window')

        # run the mainloop
        if not self.isChild:
            self._preMainLoop()
            self.top.mainloop()
            self._postMainLoop()


    def _overrideMasterSettings(self):
        """ Hook for subclasses to override some attributes if wished. """
        return


    def _preMainLoop(self):
        """ Hook for subclasses to override if wished. """
        return


    def _postMainLoop(self):
        """ Hook for subclasses to override if wished. """
        return


    def _showOpenButton(self):
        """ Should we show the "Open..." button?  Subclasses override. """
        return True


    def _setTaskParsObj(self, theTask):
        """ This method, meant to be overridden by subclasses, generates the
        _taskParsObj object. theTask can often be either a file name or a
        TaskPars subclass object. """

        # Here we catch if this version is run by accident
        raise NotImplementedError("EditParDialog is not to be used directly")


    def _saveGuiSettings(self):
        """ Hook for subclasses to save off GUI settings somewhere. """
        return # skip this by default


    def updateTitle(self, atitle):
        if atitle:
            self.top.title('%s:  %s' % (self._guiName, atitle))
        else:
            self.top.title('%s' % (self._guiName))


    def checkAllTriggers(self, action):
        """ Go over all widgets and let them know they have been edited
            recently and they need to check for any trigger actions.  This
            would be used right after all the widgets have their values
            set or forced (e.g. via setAllEntriesFromParList). """
        for entry in self.entryNo:
            entry.widgetEdited(action=action, skipDups=False)


    def freshenFocus(self):
        """ Did something which requires a new look.  Move scrollbar up.
            This often needs to be delayed a bit however, to let other
            events in the queue through first. """
        self.top.update_idletasks()
        self.top.after(10, self.setViewAtTop)


    def setViewAtTop(self):
        self.entryNo[0].focus_set()
        self.top.f.canvas.xview_moveto(0.0)
        self.top.f.canvas.yview_moveto(0.0)


    def getTaskParsObj(self):
        """ Simple accessor.  Return the _taskParsObj object. """
        return self._taskParsObj

    def mwl(self, event):
        """Mouse Wheel - under tkinter we seem to need Tk v8.5+ for this """
        if event.num == 4: # up on Linux
            self.top.f.canvas.yview_scroll(-1*self._tmwm, 'units')
        elif event.num == 5: # down on Linux
            self.top.f.canvas.yview_scroll(1*self._tmwm, 'units')
        else: # assume event.delta has the direction, but reversed sign
            self.top.f.canvas.yview_scroll(-(event.delta)*self._tmwm, 'units')

# A bug appeared in Python 2.3 that caused tk_focusNext and
# tk_focusPrev to fail. The follwoing two routines now will
# trap this error and call "fixed" versions of these tk routines
# instead in the event of such errors.

    def focusNext(self, event):
        """Set focus to next item in sequence"""
        try:
            event.widget.tk_focusNext().focus_set()
        except TypeError:
            # see tkinter equivalent code for tk_focusNext to see
            # commented original version
            name = event.widget.tk.call('tk_focusNext', event.widget._w)
            event.widget._nametowidget(str(name)).focus_set()

    def focusPrev(self, event):
        """Set focus to previous item in sequence"""
        try:
            event.widget.tk_focusPrev().focus_set()
        except TypeError:
            # see tkinter equivalent code for tk_focusPrev to see
            # commented original version
            name = event.widget.tk.call('tk_focusPrev', event.widget._w)
            event.widget._nametowidget(str(name)).focus_set()

    def doScroll(self, event):
        """Scroll the panel down to ensure widget with focus to be visible

        Tracks the last widget that doScroll was called for and ignores
        repeated calls.  That handles the case where the focus moves not
        between parameter entries but to someplace outside the hierarchy.
        In that case the scrolling is not expected.

        Returns false if the scroll is ignored, else true.
        """
        canvas = self.top.f.canvas
        widgetWithFocus = event.widget
        if widgetWithFocus is self.lastFocusWidget:
            return FALSE
        self.lastFocusWidget = widgetWithFocus
        if widgetWithFocus is None:
            return TRUE
        # determine distance of widget from top & bottom edges of canvas
        y1 = widgetWithFocus.winfo_rooty()
        y2 = y1 + widgetWithFocus.winfo_height()
        cy1 = canvas.winfo_rooty()
        cy2 = cy1 + canvas.winfo_height()
        yinc = self.yscrollincrement
        if y1<cy1:
            # this will continue to work when integer division goes away
            sdist = int((y1-cy1-yinc+1.)/yinc)
            canvas.yview_scroll(sdist, "units")
        elif cy2<y2:
            sdist = int((y2-cy2+yinc-1.)/yinc)
            canvas.yview_scroll(sdist, "units")
        return TRUE


    def _handleParListMismatch(self, probStr, extra=False):
        """ Handle the situation where two par lists do not match.
        This is meant to allow subclasses to override. Note that this only
        handles "missing" pars and "extra" pars, not wrong-type pars. """

        errmsg = 'ERROR: mismatch between default and current par lists ' + \
               'for task "'+self.taskName+'"'
        if probStr:
            errmsg += '\n\t'+probStr
        errmsg += '\n(try: "unlearn '+self.taskName+'")'
        print(errmsg)
        return False


    def _setupDefaultParamList(self):
        """ This creates self.defaultParamList.  It also does some checks
        on the paramList, sets its order if needed, and deletes any extra
        or unknown pars if found. We assume the order of self.defaultParamList
        is the correct order. """

        # Obtain the default parameter list
        self.defaultParamList = self._taskParsObj.getDefaultParList()
        theParamList = self._taskParsObj.getParList()

        # Lengths are probably equal but this isn't necessarily an error
        # here, so we check for differences below.
        if len(self.defaultParamList) != len(theParamList):
            # whoa, lengths don't match (could be some missing or some extra)
            pmsg = 'Current list not same length as default list'
            if not self._handleParListMismatch(pmsg):
                return False

        # convert current par values to a dict of { par-fullname:par-object }
        # for use below
        ourpardict = {}
        for par in theParamList: ourpardict[par.fullName()] = par

        # Sort our paramList according to the order of the defaultParamList
        # and repopulate the list according to that order. Create sortednames.
        sortednames = [p.fullName() for p in self.defaultParamList]

        # Rebuild par list sorted into correct order.  Also find/flag any
        # missing pars or any extra/unknown pars.  This automatically deletes
        # "extras" by not adding them to the sorted list in the first place.
        migrated = []
        newList = []
        for fullName in sortednames:
            if fullName in ourpardict:
                newList.append(ourpardict[fullName])
                migrated.append(fullName) # make sure all get moved over
            else: # this is a missing par - insert the default version
                theDfltVer = \
                    [p for p in self.defaultParamList if p.fullName()==fullName]
                newList.append(copy.deepcopy(theDfltVer[0]))

        # Update!  Next line writes to the self._taskParsObj.getParList() obj
        theParamList[:] = newList # fill with newList, keep same mem pointer

        # See if any got left out
        extras = [fn for fn in ourpardict if not fn in migrated]
        for fullName in extras:
            # this is an extra/unknown par - let subclass handle it
            if not self._handleParListMismatch('Unexpected par: "'+\
                        fullName+'"', extra=True):
                return False
            print('Ignoring unexpected par: "'+p+'"')

        # return value indicates that all is well to continue
        return True


    # Method to create the parameter entries
    def makeEntries(self, master, statusBar):

        # Get model data, the list of pars
        theParamList = self._taskParsObj.getParList()

        # Determine the size of the longest input string
        inputLength = INPUTWIDTH
        for i in range(self.numParams):
            inputString = theParamList[i].name
            if len(inputString) > inputLength:
                inputLength = len(inputString)

        # Set up the field widths
        # Allow extra spaces for buffer and in case the longest parameter
        # has the hidden parameter indicator
        self.fieldWidths = {}
        self.fieldWidths['inputWidth'] = inputLength + 4
        self.fieldWidths['valueWidth'] = VALUEWIDTH
        self.fieldWidths['promptWidth'] = PROMPTWIDTH

        # Loop over the parameters to create the entries
        self.entryNo = [None] * self.numParams
        dfltsVerb = self._defaultsButtonTitle
        if dfltsVerb[-1]=='s': dfltsVerb = dfltsVerb[:-1]
        for i in range(self.numParams):
            scope = theParamList[i].scope
            eparOpt = self._nonStandardEparOptionFor(theParamList[i].type)
            cbo = self._defineEditedCallbackObjectFor(scope,
                                                      theParamList[i].name)
            hcbo = None
            if self._knowTaskHelpIsHtml:
                hcbo = self
            self.entryNo[i] = eparoption.eparOptionFactory(master, statusBar,
                                  theParamList[i], self.defaultParamList[i],
                                  self.doScroll, self.fieldWidths,
                                  plugIn=eparOpt, editedCallbackObj=cbo,
                                  helpCallbackObj=hcbo, mainGuiObj=self,
                                  defaultsVerb=dfltsVerb, bg=self._entsColor,
                                  indent = scope not in (None, '', '.'),
                                  flagging = self._flagNonDefaultVals,
                                  flaggedColor=self._flagColor)


    def _nonStandardEparOptionFor(self, paramTypeStr):
        """ Hook to allow subclasses to employ their own GUI option type.
            Return None or a class which derives from EparOption. """
        return None


    def _defineEditedCallbackObjectFor(self, parScope, parName):
        """ Hook to allow subclasses to set their own callback-containing
            object to be used when a given option/parameter is edited.
            See notes in EparOption. """
        return None


    def _isUnpackagedTask(self):
        """ Hook to allow subclasses to state that this is a rogue task, not
            affiliated with a specific package, affecting its display. """
        return self.pkgName is None or len(self.pkgName) < 1


    def _toggleSectionActiveState(self, sectionName, state, skipList):
        """ Make an entire section (minus skipList items) either active or
            inactive.  sectionName is the same as the param's scope. """

        # Get model data, the list of pars
        theParamList = self._taskParsObj.getParList()

        # Loop over their assoc. entries
        for i in range(self.numParams):
            if theParamList[i].scope == sectionName:
                if skipList and theParamList[i].name in skipList:
#                   self.entryNo[i].setActiveState(True) # these always active
                    pass # if it started active, we don't need to reactivate it
                else:
                    self.entryNo[i].setActiveState(state)


    # Method to print the package and task names and to set up the menu
    # button for the choice of the display for the task help page
    def printNames(self, top, taskName, pkgName):

        topbox = Frame(top, bg=self._taskColor)
        textbox = Frame(topbox, bg=self._taskColor)
#       helpbox = Frame(topbox, bg=self._taskColor)

        # Set up the information strings
        if self._isUnpackagedTask():
            # label for a parameter list is just filename
            packString = " "+self._unpackagedTaskTitle+" = "+taskName
            Label(textbox, text=packString, bg=self._taskColor).pack(side=TOP,
                  anchor=W)
        else:
            # labels for task
            packString = "  Package = " + pkgName.upper()
            Label(textbox, text=packString, bg=self._taskColor).pack(side=TOP,
                  anchor=W)

            taskString = "       Task = " + taskName.upper()
            Label(textbox, text=taskString, bg=self._taskColor).pack(side=TOP,
                  anchor=W)
        textbox.pack(side=LEFT, anchor=W)
        topbox.pack(side=TOP, expand=FALSE, fill=X)

    # Method to set up the parent menu bar
    def makeMenuBar(self, top):

        menubar = Frame(top, bd=1, relief=GROOVE, bg=self._frmeColor)

        # Generate the menus
        fileMenu = self.makeFileMenu(menubar)

        if self._showOpenButton():
            openMenu = self.makeOpenMenu(menubar)

        # When redesigned, optionsMenu should only be on the parent
        #if not self.isChild:
        #    optionsMenu = self.makeOptionsMenu(menubar)
        optionsMenu = self.makeOptionsMenu(menubar)

        helpMenu = self.makeHelpMenu(menubar)

        menubar.pack(fill=X)


    # Method to generate a "File" menu
    def makeFileMenu(self, menubar):

        fileButton = Menubutton(menubar, text='File', bg=self._frmeColor)
        fileButton.pack(side=LEFT, padx=2)

        fileButton.menu = Menu(fileButton, tearoff=0)

#       fileButton.menu.add_command(label="Open...", command=self.pfopen)

        if self._showExecuteButton:
            fileButton.menu.add_command(label="Execute", command=self.execute)
            if self.isChild:
                fileButton.menu.entryconfigure(0, state=DISABLED)

        saqlbl ="Save"
        if self._useSimpleAutoClose: saqlbl += " & Quit"
        fileButton.menu.add_command(label=saqlbl,
                                    command=self.saveAndClose)
        if not self.isChild:
            fileButton.menu.add_command(label="Save As...", command=self.saveAs)
        fileButton.menu.add_separator()
        fileButton.menu.add_command(label=self._defaultsButtonTitle,
                                    command=self.unlearn)
        fileButton.menu.add_separator()
        if not self._useSimpleAutoClose:
            fileButton.menu.add_command(label="Close", command=self.closeGui)
        fileButton.menu.add_command(label="Cancel", command=self.abort)

        # Associate the menu with the menu button
        fileButton["menu"] = fileButton.menu

        return fileButton

    def _updateOpen(self):
        # Get new data
        flist = self._getOpenChoices()

        # Delete old choices
        if self._numOpenMenuItems > 0:
            self._openMenu.delete(0, self._numOpenMenuItems-1)

        # Add all new choices
        self._numOpenMenuItems = len(flist)
        if self._numOpenMenuItems > 0:
            for ff in flist:
                if ff[-3:] == '...':
                    self._openMenu.add_separator()
                    self._numOpenMenuItems += 1
                self._openMenu.add_radiobutton(label=ff, command=self.pfopen,
                                               variable=self._openMenuChoice,
                                               indicatoron=0)
                                               # value=ff) ... (same as label)
            self._openMenuChoice.set(0) # so nothing has check mark next to it
        else:
            showwarning(title="No Files To Open", message="No extra "+ \
                'parameter files found for task "'+self.taskName+'".')

    def _getOpenChoices(self):
        """ Get the current list of file name choices for the Open button.
            This is meant for subclasses to override. """
        return []

    # Method to generate an "Open" menu
    def makeOpenMenu(self, menubar):

        self._openMenuChoice = StringVar() # this is used till GUI closes
        self._numOpenMenuItems = 1 # see dummy

        openBtn = Menubutton(menubar, text='Open...', bg=self._frmeColor)
        openBtn.bind("<Enter>", self.printOpenInfo)
        openBtn.pack(side=LEFT, padx=2)

        openBtn.menu = Menu(openBtn, tearoff=0, postcommand=self._updateOpen)
        openBtn.menu.bind("<Enter>", self.printOpenInfo)
        openBtn.menu.add_radiobutton(label=' ', # dummy, no command
                                     variable=self._openMenuChoice)
                                     # value=fname ... (same as label)

        if self.isChild:
            openBtn.menu.entryconfigure(0, state=DISABLED)

        # Associate the menu with the menu button
        openBtn["menu"] = openBtn.menu

        # Keep a ref for ourselves
        self._openMenu = openBtn.menu

        return openBtn

    # Method to generate the "Options" menu for the parent EPAR only
    def makeOptionsMenu(self, menubar):

        # Set up the menu for the various choices they have
        self._helpChoice = StringVar()
        if self._showHelpInBrowser:
            self._helpChoice.set("BROWSER")
        else:
            self._helpChoice.set("WINDOW")

        if self._showSaveCloseOnExec:
            self._execChoice = IntVar()
            self._execChoice.set(int(self._saveAndCloseOnExec))

        optionButton = Menubutton(menubar, text="Options", bg=self._frmeColor)
        optionButton.pack(side=LEFT, padx=2)
        optionButton.menu = Menu(optionButton, tearoff=0)
        optionButton.menu.add_radiobutton(label="Display Task Help in a Window",
                                     value="WINDOW", command=self.setHelpType,
                                     variable=self._helpChoice)
        optionButton.menu.add_radiobutton(label="Display Task Help in a Browser",
                                     value="BROWSER", command=self.setHelpType,
                                     variable=self._helpChoice)

        if self._showExecuteButton and self._showSaveCloseOnExec:
            optionButton.menu.add_separator()
            optionButton.menu.add_checkbutton(label="Save and Close on Execute",
                                              command=self.setExecOpt,
                                              variable=self._execChoice)
        if self._showFlaggingChoice:
            self._flagChoice = IntVar()
            self._flagChoice.set(int(self._flagNonDefaultVals))
            optionButton.menu.add_separator()
            optionButton.menu.add_checkbutton(label="Flag Non-default Values",
                                              command=self.setFlagOpt,
                                              variable=self._flagChoice)

        # Associate the menu with the menu button
        optionButton["menu"] = optionButton.menu

        return optionButton

    def capTaskName(self):
        """ Return task name with first letter capitalized. """
        return self.taskName[:1].upper() + self.taskName[1:]

    def makeHelpMenu(self, menubar):

        button = Menubutton(menubar, text='Help', bg=self._frmeColor)
        button.bind("<Enter>", self.printHelpInfo)
        button.pack(side=RIGHT, padx=2)
        button.menu = Menu(button, tearoff=0)
        button.menu.bind("<Enter>", self.printHelpInfo)
        button.menu.add_command(label=self.capTaskName()+" Help",
                                command=self.showTaskHelp)
        button.menu.add_command(label=self._appName+" Help",
                                command=self.eparHelp)
        button.menu.add_separator()
        button.menu.add_command(label='Show '+self._appName+' Log',
                                command=self.showLogHist)
        button["menu"] = button.menu
        return button

    # Method to set up the action buttons
    # Create the buttons in an order for good navigation
    def buttonBox(self, top):

        box = Frame(top, bg=self._bboxColor, bd=1, relief=SUNKEN)

        # When the Button is exited, the information clears, and the
        # Button goes back to the nonactive color.
        top.bind("<Leave>", self.clearInfo)

        # Execute the task
        if self._showExecuteButton:
            buttonExecute = Button(box, text="Execute", bg=self._bboxColor,
                                   relief=RAISED, command=self.execute,
                                   highlightbackground=self._bboxColor)
            buttonExecute.pack(side=LEFT, padx=5, pady=7)
            buttonExecute.bind("<Enter>", self.printExecuteInfo)
            if not self._useSimpleAutoClose:
                # separate this button from the others - it's unusual
                strut = Label(box, text="", bg=self._bboxColor)
                strut.pack(side=LEFT, padx=20)

            # EXECUTE button is disabled for child windows
            if self.isChild:
                buttonExecute.configure(state=DISABLED)

        # Save the parameter settings and exit from epar
        saqlbl ="Save"
        if self._useSimpleAutoClose: saqlbl += " & Quit"
        btn = Button(box, text=saqlbl, relief=RAISED, command=self.saveAndClose,
                     bg=self._bboxColor, highlightbackground=self._bboxColor)
        btn.pack(side=LEFT, padx=5, pady=7)
        btn.bind("<Enter>", self.printSaveQuitInfo)

        # Unlearn all the parameter settings (set back to the defaults)
        buttonUnlearn = Button(box, text=self._defaultsButtonTitle,
                               relief=RAISED, command=self.unlearn,
                               bg=self._bboxColor,
                               highlightbackground=self._bboxColor)
        if self._showExtraHelpButton:
            buttonUnlearn.pack(side=LEFT, padx=5, pady=7)
        else:
            buttonUnlearn.pack(side=RIGHT, padx=5, pady=7)
        buttonUnlearn.bind("<Enter>", self.printUnlearnInfo)


        # Buttons to close versus abort this edit session.
        if not self._useSimpleAutoClose:
            buttonClose = Button(box, text="Close",
                                 relief=RAISED, command=self.closeGui,
                                 bg=self._bboxColor,
                                 highlightbackground=self._bboxColor)
            buttonClose.pack(side=LEFT, padx=5, pady=7)
            buttonClose.bind("<Enter>", self.printCloseInfo)

        buttonAbort = Button(box, text="Cancel", bg=self._bboxColor,
                             relief=RAISED, command=self.abort,
                             highlightbackground=self._bboxColor)
        buttonAbort.pack(side=LEFT, padx=5, pady=7)
        buttonAbort.bind("<Enter>", self.printAbortInfo)

        # Generate the Help button
        if self._showExtraHelpButton:
            buttonHelp = Button(box, text=self.capTaskName()+" Help",
                                relief=RAISED, command=self.showTaskHelp,
                                bg=self._bboxColor,
                                highlightbackground=self._bboxColor)
            buttonHelp.pack(side=RIGHT, padx=5, pady=7)
            buttonHelp.bind("<Enter>", self.printHelpInfo)

        # Pack
        box.pack(fill=X, expand=FALSE)

    def setExecOpt(self, event=None):
        self._saveAndCloseOnExec = bool(self._execChoice.get())

    def setFlagOpt(self, event=None):
        self._flagNonDefaultVals = bool(self._flagChoice.get())
        for entry in self.entryNo:
            entry.setIsFlagging(self._flagNonDefaultVals, True)

    def setHelpType(self, event=None):
        """ Determine which method of displaying the help pages was
        chosen by the user.  WINDOW displays in a task generated scrollable
        window.  BROWSER invokes the task's HTML help pages and displays
        in a browser. """
        self._showHelpInBrowser = bool(self._helpChoice.get() == "BROWSER")


    def eparHelp(self, event=None):     self._showAnyHelp('epar')
    def showTaskHelp(self, event=None): self._showAnyHelp('task')
    def showParamHelp(self, parName):   self._showAnyHelp('task', tag=parName)
    def showLogHist(self, event=None):  self._showAnyHelp('log')


    #
    # Define flyover help text associated with the action buttons
    #

    def clearInfo(self, event):
        self.showStatus("")

    def printHelpInfo(self, event):
        self.showStatus("Display the help page", cat=TIP)

    def printUnlearnInfo(self, event):
        self.showStatus("Set all parameter values to their default settings",
                        cat=TIP)

    def printSaveQuitInfo(self, event):
        if self._useSimpleAutoClose:
            self.showStatus("Save current entries and exit this edit session",
                            cat=TIP)
        else:
            self.showStatus("Save the current entries to "+ \
                            self._taskParsObj.getFilename(), cat=TIP)

    def printOpenInfo(self, event):
        self.showStatus(
            "Load and edit parameter values from a user-specified file",
            cat=TIP)

    def printCloseInfo(self, event):
        self.showStatus("Close this edit session.  Save first?", cat=TIP)

    def printAbortInfo(self, event):
        self.showStatus(
            "Abort this edit session, discarding any unsaved changes.",cat=TIP)

    def printExecuteInfo(self, event):
        if self._saveAndCloseOnExec:
            self.showStatus(
                 "Execute the task, and save and exit this edit session",
                 cat=TIP)
        else:
            self.showStatus("Execute the task; this window will remain open",
                            cat=TIP)


    # Process invalid input values and invoke a query dialog
    def processBadEntries(self, badEntriesList, taskname, canCancel=True):

        badEntriesString = "Task " + taskname.upper() + " --\n" \
            "Invalid values have been entered.\n\n" \
            "Parameter   Bad Value   Reset Value\n"

        for i in range (len(badEntriesList)):
            badEntriesString = badEntriesString + \
                "%15s %10s %10s\n" % (badEntriesList[i][0], \
                badEntriesList[i][1], badEntriesList[i][2])

        if canCancel:
            badEntriesString += '\n"OK" to continue using'+ \
            ' the reset values, or "Cancel" to re-enter values?\n'
        else:
            badEntriesString += \
            "\n All invalid values will return to their 'Reset Value'.\n"

        # Invoke the modal message dialog
        if canCancel:
            return askokcancel("Notice", badEntriesString)
        else:
            return showwarning("Notice", badEntriesString)


    def hasUnsavedChanges(self):
        """ Determine if there are any edits in the GUI that have not yet been
        saved (e.g. to a file).  This needs to be overridden by a subclass.
        In the meantime, just default (on the safe side) to everything being
        ready-to-save. """
        return True


    def closeGui(self, event=None):
        self.saveAndClose(askBeforeSave=True, forceClose=True)


    # SAVE/QUIT: save the parameter settings and exit epar
    def saveAndClose(self, event=None, askBeforeSave=False, forceClose=False):

        # First, see if we can/should skip the save
        doTheSave = True
        if askBeforeSave:
            if self.hasUnsavedChanges():
                doTheSave = askyesno('Save?', 'Save before closing?')
            else: # no unsaved changes, so no need to save OR even to prompt
                doTheSave = False # no need to save OR prompt

        # first save the child parameters, aborting save if
        # invalid entries were encountered
        if doTheSave and self.checkSetSaveChildren():
            return

        # Save all the entries and verify them, keeping track of the
        # invalid entries which have been reset to their original input values
        self.badEntriesList = None
        if doTheSave:
            self.badEntriesList = self.checkSetSaveEntries()
            # Note, there is a BUG here - if they hit Cancel, the save to
            # file has occurred anyway (they may not care) - need to refactor.

        # If there were invalid entries, prepare the message dialog
        if self.badEntriesList:
            ansOKCANCEL = self.processBadEntries(self.badEntriesList,
                          self.taskName)
            if not ansOKCANCEL: return

        # If there were no invalid entries or the user says OK, continue...

        # Save any GUI settings we care about.  This is a good time to do so
        # even if the window isn't closing, but especially if it is.
        self._saveGuiSettings()

        # Done saving.  Only close the window if we are running in that mode.
        if not (self._useSimpleAutoClose or forceClose):
            return

        # Remove the main epar window
        self.top.focus_set()
        self.top.withdraw()

        # If not a child window, quit the entire session
        if not self.isChild:
            self.top.destroy()
            self.top.quit()

        # Declare the global variables so they can be updated
        global CHILDX
        global CHILDY

        # Reset to the start location
        CHILDX = PARENTX
        CHILDY = PARENTY


    # OPEN: load parameter settings from a user-specified file
    def pfopen(self, event=None):
        """ Load the parameter settings from a user-specified file.  Any epar
        changes here should be coordinated with the corresponding tpar pfopen
        function. """
        raise NotImplementedError("EditParDialog is not to be used directly")


    def _getSaveAsFilter(self):
        """ Return a string to be used as the filter arg to the save file
            dialog during Save-As. Override for more specific behavior. """
        return "*.*"


    def _saveAsPreSave_Hook(self, fnameToBeUsed):
        """ Allow a subclass any specific checks right before the save. """
        return None


    def _saveAsPostSave_Hook(self, fnameToBeUsed):
        """ Allow a subclass any specific checks right after the save. """
        return None


    # SAVE AS: save the parameter settings to a user-specified file
    def saveAs(self, event=None):
        """ Save the parameter settings to a user-specified file.  Any
        changes here must be coordinated with the corresponding tpar save_as
        function. """

        self.debug('Clicked Save as...')
        # On Linux Pers..Dlg causes the cwd to change, so get a copy of current
        curdir = os.getcwd()

        # The user wishes to save to a different name
        writeProtChoice = self._writeProtectOnSaveAs
        if capable.OF_TKFD_IN_EPAR:
            # Prompt using native looking dialog
            fname = asksaveasfilename(parent=self.top,
                    title='Save Parameter File As',
                    defaultextension=self._defSaveAsExt,
                    initialdir=os.path.dirname(self._getSaveAsFilter()))
        else:
            # Prompt. (could use tkinter's FileDialog, but this one is prettier)
            # initWProtState is only used in the 1st call of a session
            from . import filedlg
            fd = filedlg.PersistSaveFileDialog(self.top,
                         "Save Parameter File As", self._getSaveAsFilter(),
                         initWProtState=writeProtChoice)
            if fd.Show() != 1:
                fd.DialogCleanup()
                os.chdir(curdir) # in case file dlg moved us
                return
            fname = fd.GetFileName()
            writeProtChoice = fd.GetWriteProtectChoice()
            fd.DialogCleanup()

        if not fname: return # canceled

        # First check the child parameters, aborting save if
        # invalid entries were encountered
        if self.checkSetSaveChildren():
            os.chdir(curdir) # in case file dlg moved us
            return

        # Run any subclass-specific steps right before the save
        self._saveAsPreSave_Hook(fname)

        # Verify all the entries (without save), keeping track of the invalid
        # entries which have been reset to their original input values
        self.badEntriesList = self.checkSetSaveEntries(doSave=False)

        # If there were invalid entries, prepare the message dialog
        if self.badEntriesList:
            ansOKCANCEL = self.processBadEntries(self.badEntriesList,
                          self.taskName)
            if not ansOKCANCEL:
                os.chdir(curdir) # in case file dlg moved us
                return

        # If there were no invalid entries or the user says OK, finally
        # save to their stated file.  Since we have already processed the
        # bad entries, there should be none returned.
        mstr = "TASKMETA: task="+self.taskName+" package="+self.pkgName
        if self.checkSetSaveEntries(doSave=True, filename=fname, comment=mstr,
                                    set_ro=writeProtChoice,
                                    overwriteRO=True):
            os.chdir(curdir) # in case file dlg moved us
            raise Exception("Unexpected bad entries for: "+self.taskName)

        # Run any subclass-specific steps right after the save
        self._saveAsPostSave_Hook(fname)

        os.chdir(curdir) # in case file dlg moved us


    # EXECUTE: save the parameter settings and run the task
    def execute(self, event=None):

        self.debug('Clicked Execute')
        # first save the child parameters, aborting save if
        # invalid entries were encountered
        if self.checkSetSaveChildren():
            return

        # If we are only executing (no save and close) do so here and return
        if not self._saveAndCloseOnExec:
            # First check the parameter values
            self.badEntriesList = self.checkSetSaveEntries(doSave=False)
            # If there were invalid entries, show the message dialog
            if self.badEntriesList:
                ansOKCANCEL = self.processBadEntries(self.badEntriesList,
                              self.taskName)
                if not ansOKCANCEL: return
            self.showStatus("Task "+self.taskName+" is running...", keep=2)
            self._executed = True # note for later use
            self.runTask()
            return

        # Now save the parameter values of the parent
        self.badEntriesList = self.checkSetSaveEntries()

        # If there were invalid entries in the parent epar dialog, prepare
        # the message dialog
        if self.badEntriesList:
            ansOKCANCEL = self.processBadEntries(self.badEntriesList,
                          self.taskName)
            if not ansOKCANCEL: return

        # If there were no invalid entries or the user said OK

        # Save any GUI settings we care about since window is closing
        self._saveGuiSettings()

        # Remove the main epar window
        self.top.focus_set()
        self.top.withdraw()
        self.top.destroy()

        print("\nTask "+self.taskName+" is running...\n")

        # Before running the task, clear any already-handled exceptions that
        # will be erroneously picked up by the task's logger utility.
        # This is temporary.  Remove this line when logging is fixed.
        try:
            sys.exc_clear() # not present in PY3K
        except AttributeError:
            pass

        # Run the task
        try:
            self._executed = True # note for later use
            self.runTask()
        finally:
            self.top.quit()

        # Declare the global variables so they can be updated
        global CHILDX
        global CHILDY

        # Reset to the start location
        CHILDX = PARENTX
        CHILDY = PARENTY


    # ABORT: abort this epar session
    def abort(self, event=None):

        # Declare the global variables so they can be updated
        global CHILDX
        global CHILDY

        # Reset to the start location
        CHILDX = PARENTX
        CHILDY = PARENTY

        # Give focus back to parent window and abort
        self.top.focus_set()
        self.top.withdraw()

        self._canceled = True # note for later use

        # Do not destroy the window, just hide it for now.
        # This is so EXECUTE will not get an error - properly use Mediator.
        #self.top.destroy()
        if not self.isChild:
            self.top.destroy()
            self.top.quit()


    # UNLEARN: unlearn all the parameters by setting their values
    # back to the system default
    def unlearn(self, event=None):

        self.debug('Clicked Unlearn')
        # Reset the values of the parameters
        self.unlearnAllEntries(self.top.f.canvas.entries)
        self.freshenFocus()


    # HTMLHELP: invoke the HTML help
    def htmlHelp(self, helpString=None, title=None, istask=False, tag=None):
        """ Pop up the help in a browser window.  By default, this tries to
        show the help for the current task.  With the option arguments, it can
        be used to show any help string. """
        # Check the help string.  If it turns out to be a URL, launch that,
        # if not, dump it to a quick and dirty tmp html file to make it
        # presentable, and pass that file name as the URL.
        if not helpString:
            helpString = self.getHelpString(self.pkgName+'.'+self.taskName)
        if not title:
            title = self.taskName
        lwr = helpString.lower()
        if lwr.startswith("http:") or lwr.startswith("https:") or \
           lwr.startswith("file:"):
            url = helpString
            if tag and url.find('#') < 0:
                url += '#'+tag
#           print('LAUNCHING: '+url) # DBG
            irafutils.launchBrowser(url, subj=title)
        else:
            # Write it to a temp HTML file to display
            (fd, fname) = tempfile.mkstemp(suffix='.html', prefix='editpar_')
            os.close(fd)
            f = open(fname, 'w')
            if istask and self._knowTaskHelpIsHtml:
                f.write(helpString)
            else:
                f.write('<html><head><title>'+title+'</title></head>\n')
                f.write('<body><h3>'+title+'</h3>\n')
                f.write('<pre>\n'+helpString+'\n</pre></body></html>')
            f.close()
            irafutils.launchBrowser("file://"+fname, subj=title)


    def _showAnyHelp(self, kind, tag=None):
        """ Invoke task/epar/etc. help and put the page in a window.
        This same logic is used for GUI help, task help, log msgs, etc. """

        # sanity check
        assert kind in ('epar', 'task', 'log'), 'Unknown help kind: '+str(kind)

        #-----------------------------------------
        # See if they'd like to view in a browser
        #-----------------------------------------
        if self._showHelpInBrowser or (kind == 'task' and
                                       self._knowTaskHelpIsHtml):
            if kind == 'epar':
                self.htmlHelp(helpString=self._appHelpString,
                              title='Parameter Editor Help')
            if kind == 'task':
                self.htmlHelp(istask=True, tag=tag)
            if kind == 'log':
                self.htmlHelp(helpString='\n'.join(self._msgHistory),
                              title=self._appName+' Event Log')
            return

        #-----------------------------------------
        # Now try to pop up the regular Tk window
        #-----------------------------------------
        wins = {'epar':self.eparHelpWin,
                'task':self.irafHelpWin,
                'log': self.logHistWin, }
        window = wins[kind]
        try:
            if window.state() != NORMAL:
                window.deiconify()
            window.tkraise()
            return
        except (AttributeError, TclError):
            pass

        #---------------------------------------------------------
        # That didn't succeed (window is still None), so build it
        #---------------------------------------------------------
        if kind == 'epar':
            self.eparHelpWin = self.makeHelpWin(self._appHelpString,
                                                title='Parameter Editor Help')
        if kind == 'task':
            # Acquire the task help as a string
            # Need to include the package name for the task to
            # avoid name conflicts with tasks from other packages. WJH
            self.irafHelpWin = self.makeHelpWin(self.getHelpString(
                                                self.pkgName+'.'+self.taskName))
        if kind == 'log':
            self.logHistWin = self.makeHelpWin('\n'.join(self._msgHistory),
                                               title=self._appName+' Event Log')


    def canceled(self):
        """ Did the user click Cancel? (or close us via the window manager) """
        return self._canceled


    def executed(self):
        """ Did the user click Execute? """
        return self._executed


    # Get the task help in a string
    def getHelpString(self, taskname):
        """ Provide a task-specific help string. """
        return self._taskParsObj.getHelpAsString()


    # Set up the help dialog (browser)
    def makeHelpWin(self, helpString, title="Parameter Editor Help Browser"):

        # Generate a new Toplevel window for the browser
        # hb = Toplevel(self.top, bg="SlateGray3")
        hb = Toplevel(self.top, bg=None)
        hb.title(title)
        hb.iconLabel = title

        # Set up the Menu Bar
        hb.menubar = Frame(hb, relief=RIDGE, borderwidth=0)
        hb.menubar.button = Button(hb.menubar, text="Close",
                                     relief=RAISED,
                                     command=hb.destroy)
        hb.menubar.button.pack()
        hb.menubar.pack(side=BOTTOM, padx=5, pady=5)

        # Define the Frame for the scrolling Listbox
        hb.frame = Frame(hb, relief=RIDGE, borderwidth=1)

        # Attach a vertical Scrollbar to the Frame
        hb.frame.vscroll = Scrollbar(hb.frame, orient=VERTICAL,
                 width=11, relief=SUNKEN, activerelief=RAISED,
                 takefocus=FALSE)

        # Define the Listbox and setup the Scrollbar
        hb.frame.list = Listbox(hb.frame,
                                relief=FLAT,
                                height=25,
                                width=80,
                                takefocus=FALSE,
                                selectmode=SINGLE,
                                selectborderwidth=0)
        hb.frame.list['yscrollcommand'] = hb.frame.vscroll.set

        hb.frame.vscroll['command'] = hb.frame.list.yview
        hb.frame.vscroll.pack(side=RIGHT, fill=Y)
        hb.frame.list.pack(side=TOP, expand=TRUE, fill=BOTH)
        hb.frame.pack(side=TOP, fill=BOTH, expand=TRUE)

        # Insert each line of the helpString onto the Frame
        listing = helpString.split('\n')
        for line in listing:

            # Filter the text *** DO THIS A BETTER WAY ***
            line = line.replace("\x0e", "")
            line = line.replace("\x0f", "")
            line = line.replace("\f", "")

            # Insert the text into the Listbox
            hb.frame.list.insert(END, line)

        # When the Listbox appears, the listing will be at the beginning
        y = hb.frame.vscroll.get()[0]
        hb.frame.list.yview(int(y))

        # enable Page Up/Down keys
        scroll = hb.frame.list.yview_scroll
        hb.bind('<Next>', lambda event, fs=scroll: fs(1, "pages"))
        hb.bind('<Prior>', lambda event, fs=scroll: fs(-1, "pages"))

        # Position this dialog relative to the parent
        hb.geometry("+%d+%d" % (self.top.winfo_rootx() + HELPX,
                                     self.top.winfo_rooty() + HELPY))
        return hb

    def validate(self):

        return 1


    def setAllEntriesFromParList(self, aParList, updateModel=False):
        """ Set all the parameter entry values in the GUI to the values
            in the given par list. If 'updateModel' is True, the internal
            param list will be updated to the new values as well as the GUI
            entries (slower and not always necessary). Note the
            corresponding TparDisplay method. """

        # Get model data, the list of pars
        theParamList = self._taskParsObj.getParList() # we may modify members

        if len(aParList) != len(theParamList):
            showwarning(message="Attempting to set parameter values from a "+ \
                        "list of different length ("+str(len(aParList))+ \
                        ") than the number shown here ("+ \
                        str(len(theParamList))+").  Be aware.",
                        title="Parameter List Length Mismatch")

        # LOOP THRU GUI PAR LIST
        for i in range(self.numParams):
            par = theParamList[i]
            if par.type == "pset":
                continue # skip PSET's for now
            gui_entry = self.entryNo[i]

            # Set the value in the paramList before setting it in the GUI
            # This may be in the form of a list, or an IrafParList (getValue)
            if isinstance(aParList, list):
                # Since "aParList" can have them in different order and number
                # than we do, we'll have to first find the matching param.
                found = False
                for newpar in aParList:
                    if newpar.name==par.name and newpar.scope==par.scope:
                        par.set(newpar.value) # same as .get(native=1,prompt=0)
                        found = True
                        break

                # Now see if newpar was found in our list
                if not found:
                    pnm = par.name
                    if len(par.scope): pnm = par.scope+'.'+par.name
                    raise UnfoundParamError('Error - Unfound Parameter! \n\n'+\
                      'Expected parameter "'+pnm+'" for task "'+ \
                      self.taskName+'". \nThere may be others...')

            else: # assume has getValue()
                par.set(aParList.getValue(par.name, native=1, prompt=0))

            # gui holds a str, but par.value is native; conversion occurs
            gui_entry.forceValue(par.value, noteEdited=False) # no triggers yet

        if updateModel:
            # Update the model values via checkSetSaveEntries
            self.badEntriesList = self.checkSetSaveEntries(doSave=False)

            # If there were invalid entries, prepare the message dialog
            if self.badEntriesList:
                self.processBadEntries(self.badEntriesList,
                                       self.taskName, canCancel=False)


    def unlearnAllEntries(self, master):
        """ Method to "unlearn" all the parameter entry values in the GUI
            and set the parameter back to the default value """
        for entry in self.entryNo:
            entry.unlearnValue()


    def getValue(self, name, scope=None, native=False):
        """ Return current par value from the GUI. This does not do any
        validation, and it it not necessarily the same value saved in the
        model, which is always behind the GUI setting, in time. This is NOT
        to be used to get all the values - it would not be efficient. """

        # Get model data, the list of pars
        theParamList = self._taskParsObj.getParList()

        # NOTE: If par scope is given, it will be used, otherwise it is
        # assumed to be unneeded and the first name-match is returned.
        fullName = basicpar.makeFullName(scope, name)

        # Loop over the parameters to find the requested par
        for i in range(self.numParams):
            par = theParamList[i] # IrafPar or subclass
            entry = self.entryNo[i] # EparOption or subclass
            if par.fullName() == fullName or \
               (scope is None and par.name == name):
                if native:
                    return entry.convertToNative(entry.choice.get())
                else:
                    return entry.choice.get()
        # We didn't find the requested par
        raise RuntimeError('Could not find par: "'+fullName+'"')


    # Read, save, and validate the entries
    def checkSetSaveEntries(self, doSave=True, filename=None, comment=None,
                            fleeOnBadVals=False, allowGuiChanges=True,
                            set_ro=False, overwriteRO=False):

        self.badEntries = []
        asNative = self._taskParsObj.knowAsNative()

        # Get model data, the list of pars
        theParamList = self._taskParsObj.getParList()

        # Loop over the parameters to obtain the modified information
        for i in range(self.numParams):

            par = theParamList[i] # IrafPar or subclass
            entry = self.entryNo[i] # EparOption or subclass
            # Cannot change an entry if it is a PSET, just skip
            if par.type == "pset":
                continue

            # get current state of par in the gui
            value = entry.choice.get()

            # Set new values for changed parameters - a bit tricky,
            # since changes that weren't followed by a return or
            # tab have not yet been checked.  If we eventually
            # use a widget that can check all changes, we will
            # only need to check the isChanged flag.
            if par.isChanged() or value != entry.previousValue:

                # CHECK: Verify the value. If its invalid (and allowGuiChanges),
                # the value will be converted to its original valid value.
                # Maintain a list of the reset values for user notification.
                # Always call entryCheck, no matter what type of _taskParsObj,
                # since entryCheck can do some basic type checking.
                failed = False
                if entry.entryCheck(repair=allowGuiChanges):
                    failed = True
                    self.badEntries.append([entry.name, value,
                                           entry.choice.get()])
                    if fleeOnBadVals: return self.badEntries
                # See if we need to do a more serious validity check
                elif self._taskParsObj.canPerformValidation():
                    # if we are planning to save in native type, test that way
                    if asNative:
                        try:
                            value = entry.convertToNative(value)
                        except:
                            failed = True
                            prev = entry.previousValue
                            self.badEntries.append([entry.name, value, prev])
                            if fleeOnBadVals: return self.badEntries
                            if allowGuiChanges: entry.choice.set(prev)
                    # now try the val in it's validator
                    if not failed:
                        valOK, prev = self._taskParsObj.tryValue(entry.name,
                                                        value, scope=par.scope)
                        if not valOK:
                            failed = True
                            self.badEntries.append([entry.name,str(value),prev])
                            if fleeOnBadVals: return self.badEntries
                            if allowGuiChanges: entry.choice.set(prev)

                # get value again in case it changed - this version IS valid
                value = entry.choice.get()
                if asNative: value = entry.convertToNative(value)

                # SET: Update the task parameter (also does the conversion
                # from string)
                self._taskParsObj.setParam(par.name, value, scope=par.scope,
                                           check=0, idxHint=i)

        # SAVE: Save results to the given file
        if doSave:
            self.debug('Saving...')
            out = self._doActualSave(filename, comment, set_ro=set_ro,
                                     overwriteRO=overwriteRO)
            if len(out):
                self.showStatus(out, keep=2) # inform user on saves

        return self.badEntries


    def _doActualSave(self, fname, comment, set_ro=False):
        """ Here we call the method on the _taskParsObj to do the actual
        save.  Return a string result to be printed to the screen. """
        # do something like
#       return self._taskParsObj.saveParList(filename=fname, comment=comment)
        raise NotImplementedError("EditParDialog is not to be used directly")


    def checkSetSaveChildren(self, doSave=True):
        """Check, then set, then save the parameter settings for
        all child (pset) windows.

        Prompts if any problems are found.  Returns None
        on success, list of bad entries on failure.
        """
        if self.isChild:
            return

        # Need to get all the entries and verify them.
        # Save the children in backwards order to coincide with the
        # display of the dialogs (LIFO)
        for n in range (len(self.top.childList)-1, -1, -1):
            self.badEntriesList = self.top.childList[n]. \
                                  checkSetSaveEntries(doSave=doSave)
            if self.badEntriesList:
                ansOKCANCEL = self.processBadEntries(self.badEntriesList,
                              self.top.childList[n].taskName)
                if not ansOKCANCEL:
                    return self.badEntriesList
            # If there were no invalid entries or the user says OK,
            # close down the child and increment to the next child
            self.top.childList[n].top.focus_set()
            self.top.childList[n].top.withdraw()
            del self.top.childList[n]
        # all windows saved successfully
        return


    def _pushMessages(self):
        """ Internal callback used to make sure the msg list keeps moving. """
        # This continues to get itself called until no msgs are left in list.
        self.showStatus('')
        if len(self._statusMsgsToShow) > 0:
            self.top.after(200, self._pushMessages)


    def debug(self, msg):
        """ Convenience function.  Use showStatus without puting into GUI. """
        self.showStatus(msg, cat=DBG)


    def showStatus(self, msg, keep=0, cat=None):
        """ Show the given status string, but not until any given delay from
            the previous message has expired. keep is a time (secs) to force
            the message to remain without being overwritten or cleared. cat
            is a string category used only in the historical log. """
        # prep it, space-wise
        msg = msg.strip()
        if len(msg) > 0:
            # right here is the ideal place to collect a history of messages
            forhist = msg
            if cat: forhist = '['+cat+'] '+msg
            forhist = time.strftime("%a %H:%M:%S")+': '+forhist
            self._msgHistory.append(forhist)
            # now set the spacing
            msg = '  '+msg

        # stop here if it is a category not shown in the GUI
        if cat == DBG:
            return

        # see if we can show it
        now = time.time()
        if now >= self._leaveStatusMsgUntil: # we are clear, can show a msg
            # first see if this msg is '' - if so we will show an important
            # waiting msg instead of the '', and then pop it off our list
            if len(msg) < 1 and len(self._statusMsgsToShow) > 0:
                msg, keep = self._statusMsgsToShow[0] # overwrite both args
                del self._statusMsgsToShow[0]
            # now actuall print the status out to the status widget
            self.top.status.config(text = msg)
            # reset our delay flag
            self._leaveStatusMsgUntil = 0
            if keep > 0:
                self._leaveStatusMsgUntil = now + keep
        else:
            # there is a previous message still up, is this one important?
            if len(msg) > 0 and keep > 0:
                # Uh-oh, this is an important message that we don't want to
                # simply skip, but on the other hand we can't show it yet...
                # So we add it to _statusMsgsToShow and show it later (asap)
                if (msg,keep) not in self._statusMsgsToShow:
                    if len(self._statusMsgsToShow) < 7:
                        self._statusMsgsToShow.append( (msg,keep) ) # tuple
                        # kick off timer loop to get this one pushed through
                        if len(self._statusMsgsToShow) == 1:
                            self._pushMessages()
                    else:
                        # should never happen, but just in case
                        print("Lost message!: "+msg+" (too far behind...)")

    # Run the task
    def runTask(self):

        # Use the run method of the IrafTask class
        # Set mode='h' so it does not prompt for parameters (like IRAF epar)
        # Also turn on parameter saving
        try:
            self._taskParsObj.run(mode='h', _save=1)
        except taskpars.NoExecError as nee:  # catch only this, let all else thru
            showwarning(message="No way found to run task\n\n"+\
                        str(nee), title="Can Not Run Task")



from __future__ import absolute_import, print_function, division

import multiprocessing
import time

from ..mputil import launch_and_wait, best_tile_layout


def takes_time(x):
    """Example function which takes some time to run."""
    time.sleep(0.001)  # 1 ms is long by computer standards?


def test_launch_and_wait():
    """Illustrate use of launch_and_wait"""
    p = None
    subprocs = []

    for item in range(2, 10):
        # print(("mputil: instantiating Process for x = "+str(item)))
        p = multiprocessing.Process(target=takes_time, args=(item,),
                                    name='takes_time()')
        subprocs.append(p)

    # launch em, pool-fashion
    launch_and_wait(subprocs, 3)

    # NOTE: If no exception raised by now, considered passed.
    # by now, all should be finished
    # print("All subprocs should be finished and joined.")


def test_best_tile_layout():
    """Loop though some numbers and make sure we get expected results."""
    for i in range(257):
        x, y = best_tile_layout(i)
        assert (x * y <= i) or (i == 0), \
            "Total num resulting tiles > pool_size"
        unused_cores = i - (x * y)
        # print(i, (x,y), unused_cores)
        if i < 10:
            assert unused_cores <= 1, "Too many idle cores at i = " + str(i)
        else:
            percent_unused = 100. * ((unused_cores * 1.) / i)
            assert percent_unused < 14., "Too many idles cores at i: " + str(i)



from __future__ import absolute_import

import pytest
from ..irafutils import csvSplit


@pytest.mark.parametrize('input_str,input_len,expected', [
    (None, 0, "[]"),
    ('', 0, "[]"),
    (' ', 1, "[' ']"),
    ('a', 1, "['a']"),
    (',', 2, "['', '']"),
    (',a', 2, "['', 'a']"),
    ('a,', 2, "['a', '']"),
    (',a,', 3, "['', 'a', '']"),
    ("abc'-hi,ya-'xyz", 1, """["abc'-hi,ya-'xyz"]"""),
    ('abc"double-quote,eg"xy,z', 2, """['abc"double-quote,eg"xy', 'z']"""),
    ('abc"""triple-quote,eg"""xyz', 1, '[\'abc"""triple-quote,eg"""xyz\']'),
    ("'s1', 'has, comma', z", 3, """["'s1'", " 'has, comma'", ' z']"""),
    ("a='s1', b='has,comma,s', c", 3, """["a='s1'", " b='has,comma,s'", ' c']"""),
])
def test_csvSplit(input_str, input_len, expected):
    result = csvSplit(input_str, ',', True)
    result_len = len(result)

    assert result_len == input_len and repr(result) == expected, \
        "For case: {} expected:\n{}\nbut got:\n{}".format(
            input_str, expected, repr(result))



""" This is a temporary module, used during (and for a while after) the
transition to Python 3.  This code is planned to be kept in place until
the least version of Python supported no longer requires it (and of course
until all callers no longer need it).
This code should run as-is in 2.x and also run unedited after 2to3 in 3.x.

$Id$
"""

from __future__ import division # confidence high
import os, sys
PY3K = sys.version_info[0] > 2


def ndarr2str(arr, encoding='ascii'):
    """ This is used to ensure that the return value of arr.tostring()
    is actually a string.  This will prevent lots of if-checks in calling
    code.  As of numpy v1.6.1 (in Python 3.2.3), the tostring() function
    still returns type 'bytes', not 'str' as it advertises. """
    # be fast, don't check - just assume 'arr' is a numpy array - the tostring
    # call will fail anyway if not
    retval = arr.tostring()
    # would rather check "if isinstance(retval, bytes)", but support 2.5.
    # could rm the if PY3K check, but it makes this faster on 2.x.
    if PY3K and not isinstance(retval, str):
        return retval.decode(encoding)
    else: # is str
        return retval


def ndarr2bytes(arr, encoding='ascii'):
    """ This is used to ensure that the return value of arr.tostring()
    is actually a *bytes* array in PY3K.  See notes in ndarr2str above.  Even
    though we consider it a bug that numpy's tostring() function returns
    a bytes array in PY3K, there are actually many instances where that is what
    we want - bytes, not unicode.  So we use this function in those
    instances to ensure that when/if this numpy "bug" is "fixed", that
    our calling code still gets bytes where it needs/expects them. """
    # be fast, don't check - just assume 'arr' is a numpy array - the tostring
    # call will fail anyway if not
    retval = arr.tostring()
    # would rather check "if not isinstance(retval, bytes)", but support 2.5.
    if PY3K and isinstance(retval, str):
        # Take note if this ever gets used.  If this ever occurs, it
        # is likely wildly inefficient since numpy.tostring() is now
        # returning unicode and numpy surely has a tobytes() func by now.
        # If so, add a code path to call its tobytes() func at our start.
        return retval.encode(encoding)
    else: # is str==bytes in 2.x
        return retval


def tobytes(s, encoding='ascii'):
    """ Convert string s to the 'bytes' type, in all Pythons, even
    back before Python 2.6.  What 'str' means varies by PY3K or not.
    In Pythons before 3.0, this is technically the same as the str type
    in terms of the character data in memory. """
    # NOTE: after we abandon 2.5, we might simply instead use "bytes(s)"
    # NOTE: after we abandon all 2.*, del this and prepend byte strings with 'b'
    if PY3K:
        if isinstance(s, bytes):
            return s
        else:
            return s.encode(encoding)
    else:
        # for py2.6 on (before 3.0), bytes is same as str;  2.5 has no bytes
        # but handle if unicode is passed
        if isinstance(s, unicode):
            return s.encode(encoding)
        else:
            return s

def tostr(s, encoding='ascii'):
    """ Convert string-like-thing s to the 'str' type, in all Pythons, even
    back before Python 2.6.  What 'str' means varies by PY3K or not.
    In Pythons before 3.0, str and bytes are the same type.
    In Python 3+, this may require a decoding step. """
    if PY3K:
        if isinstance(s, str): # str == unicode in PY3K
            return s
        else: # s is type bytes
            return s.decode(encoding)
    else:
        # for py2.6 on (before 3.0), bytes is same as str;  2.5 has no bytes
        # but handle if unicode is passed
        if isinstance(s, unicode):
            return s.encode(encoding)
        else:
            return s


try:
    BNULLSTR = tobytes('')   # after dropping 2.5, change to: b''
    BNEWLINE = tobytes('\n') # after dropping 2.5, change to: b'\n'
except:
    BNULLSTR = ''
    BNEWLINE = '\n'


def bytes_read(fd, sz):
   """ Perform an os.read in a way that can handle both Python2 and Python3
   IO.  Assume we are always piping only ASCII characters (since that is all
   we have ever done with IRAF).  Either way, return the data as bytes.
   """
#  return tobytes(os.read(fd, sz))
   return os.read(fd, sz) # already returns str in Py2.x and bytes in PY3K


def bytes_write(fd, bufstr):
   """ Perform an os.write in a way that can handle both Python2 and Python3
   IO.  Assume we are always piping only ASCII characters (since that is all
   we have ever done with IRAF).  Either way, write the binary data to fd.
   """
   return os.write(fd, tobytes(bufstr))



####
#       Class Dialog
#
#       Purpose
#       Base class for many dialog box classes.
####
"""
$Id$
"""
from __future__ import division # confidence high

import sys

PY3K = sys.version_info[0] > 2
if PY3K:
    from tkinter import *
else:
    from Tkinter import *

class Dialog:

    def __init__(self, master):
        self.master = master
        self.top = Toplevel(self.master)
        self.top.title(self.__class__.__name__)
        self.top.minsize(1, 1)
        self.myWaitVar = str(self.top) + 'EndDialogVar'

    def Show(self):
        self.SetupDialog()
        self.CenterDialog()
        self.top.deiconify()
        self.top.focus()

    def TerminateDialog(self, withValue):
        self.top.setvar(self.myWaitVar, withValue)
        self.top.withdraw()

    def DialogCleanup(self):
        self.top.destroy()
        self.master.focus()

    def SetupDialog(self):
        pass

    def CenterDialog(self):
        self.top.withdraw()
        self.top.update_idletasks()
        w = self.top.winfo_screenwidth()
        h = self.top.winfo_screenheight()
        reqw = self.top.winfo_reqwidth()
        reqh = self.top.winfo_reqheight()
        centerx = str((w-reqw)//2)
        centery = str((h-reqh)//2 - 100)
        geomStr = "+" + centerx + "+" + centery
        self.top.geometry(geomStr)

####
#       Class ModalDialog
#
#       Purpose
#       Base class for many modal dialog box classes.
####

class ModalDialog(Dialog):

    def __init__(self, master):
        Dialog__init__(self, master)

    def Show(self):
        self.SetupDialog()
        self.CenterDialog()
        try:
            self.top.grab_set() # make it modal
        except TclError:
            # This fails on Linux, but does it really HAVE to be modal
            if sys.platform.lower().find('linux') >= 0:
                pass
            else:
                raise
        self.top.focus()
        self.top.deiconify()
        self.top.waitvar(self.myWaitVar)
        return int(self.top.getvar(self.myWaitVar))



#!/usr/bin/env python
"""
        convertlog: Read ASCII trailer file and convert it to a waivered-FITS file.

        License: http://www.stsci.edu/resources/software_hardware/pyraf/LICENSE

        Usage:

                convertlog.py [OPTIONS] trailer_filename

        :Options:

        -h         print the help (this text)

        -v         print version of task

        -w
        --width    Width (in chars) for trailer file table column

        -o
        --output   Name of output FITS trailer file
                   If none is specified, it will convert input file
                   from "rootname.tra" to "rootname_trl.fits"

        :Example:

        If used in Pythons script, a user can, e. g.::

            >>> from stsci.tools import convertlog
            >>> convertlog.convert(TRLFileName)  # doctest: +SKIP

        The most basic usage from the command line::

            convertlog test1.tra

        This command will convert the input ASCII trailer file test1.tra to
        a waivered-FITS file test1_trl.fits.

"""
# Developed by Science Software Branch, STScI, USA.

from __future__ import division, print_function # confidence high

__version__ = "1.0 (7 Jan, 2016), \xa9 AURA"

import os
import sys
from astropy.io import fits
import numpy as np

import textwrap

def convert(input, width=132, output=None, keep=False):

    """Input ASCII trailer file "input" will be read.

    The contents will then be written out to a FITS file in the same format
    as used by 'stwfits' from IRAF.

    Parameters
    ===========
    input : str
        Filename of input ASCII trailer file

    width : int
        Number of characters wide to use for defining output FITS column
        [Default: 132]

    output : str
        Filename to use for writing out converted FITS trailer file
        If None, input filename will be converted from *.tra -> *_trl.fits
        [Default: None]

    keep : bool
        Specifies whether or not to keep any previously written FITS files
        [Default: False]

    """
    # open input trailer file
    trl = open(input)

    # process all lines
    lines = np.array([i for text in trl.readlines() for i in textwrap.wrap(text,width=width)])

    # close ASCII trailer file now that we have processed all the lines
    trl.close()

    if output is None:
        # create fits file
        rootname,suffix = os.path.splitext(input)
        s = suffix[1:].replace('ra','rl')
        fitsname = "{}_{}{}fits".format(rootname,s,os.path.extsep)
    else:
        fitsname = output
    full_name = os.path.abspath(os.path.join(os.path.curdir,fitsname))

    old_file = os.path.exists(full_name)
    if old_file:
        if keep:
            print("ERROR: Trailer file already written out as: {}".format(full_name))
            raise IOError
        else:
            os.remove(full_name)

    # Build FITS table and write it out
    line_fmt = "{}A".format(width)
    tbhdu = fits.BinTableHDU.from_columns([fits.Column(name='TEXT_FILE',format=line_fmt,array=lines)])
    tbhdu.writeto(fitsname)

    print("Created output FITS filename for trailer:{}    {}".format(os.linesep,full_name))

    os.remove(input)

def usage():
    print(__doc__)

def main():
    import getopt

    try:
        optlist, args = getopt.getopt(sys.argv[1:], 'hvkw:o:')
    except getopt.error as e:
        print(str(e))
        print(__doc__)
        print("\t", __version__)
        sys.exit(2)

    output = None
    width = 132
    keep = False

    for o, a in optlist:
        if o in ("-h", "--help"):
            usage()
            sys.exit()
        elif o in ("-o", "--output"):
            output = a
        elif o in ("-w", "--width"):
            width = int(a)
        elif o in ("-k", "--keep"):
            keep = True
        else:
            assert False, "unhandled option"


    trl_file = args[0]
    try:
        print("Converting {}...".format(trl_file))
        convert(trl_file, width=width, output=output,keep=keep)
    except:
        print("ERROR: Convertlog failed to convert: {}".format(trl_file))
        sys.exit(2)

#-------------------------------------------------------------------------------
# special initialization when this is the main program

if __name__ == "__main__":
    main()

"""

Copyright (C) 2003 Association of Universities for Research in Astronomy (AURA)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    3. The name of AURA and its representatives may not be used to
      endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
"""



"""teal_bttn.py: for defining the action "parameter" button widget
   to be used in TEAL.

$Id$
"""
from __future__ import division, print_function # confidence high

import traceback
from . import eparoption, vtor_checks

class TealActionParButton(eparoption.ActionEparButton):

    def getButtonLabel(self):
        """ Return string to be used on as button label - "value" of par. """
        # If the value has a comma, return the 2nd part, else use whole thing
        return self.value.split(',')[-1].strip()

    def getShowName(self):
        """ Return string to be used on LHS of button - "name" of par. """
        # If the value has a comma, return the 1st part, else leave empty
        if self.value.find(',') >= 0:
            return self.value.split(',')[0]
        else:
            return ''

    def flagThisPar(self, currentVal, force):
        """ Override this to do nothing - the value of this par will
        never be wrong and thus never need to be flagged. """
        pass

    def clicked(self):
        """ Called when this button is clicked. Execute code from .cfgspc """
        try:
            from . import teal
        except:
            teal = None
        try:
            # start drilling down into the tpo to get the code
            tealGui = self._mainGuiObj
            tealGui.showStatus('Clicked "'+self.getButtonLabel()+'"', keep=1)
            pscope = self.paramInfo.scope
            pname = self.paramInfo.name
            tpo = tealGui._taskParsObj
            tup = tpo.getExecuteStrings(pscope, pname)
            code = ''
            if not tup:
                if teal:
                    teal.popUpErr(tealGui.top, "No action to perform",
                                  "Action Button Error")
                return
            for exname in tup:
                if '_RULES_' in tpo and exname in tpo['_RULES_'].configspec:
                    ruleSig = tpo['_RULES_'].configspec[exname]
                    chkArgsDict = vtor_checks.sigStrToKwArgsDict(ruleSig)
                    code = chkArgsDict.get('code') # a string or None
                    # now go ahead and execute it
                    teal.execEmbCode(pscope, pname, self.getButtonLabel(),
                                     tealGui, code)
            # done
            tealGui.debug('Finished: "'+self.getButtonLabel()+'"')
        except Exception as ex:
            msg = 'Error executing: "'+self.getButtonLabel()+'"\n'+ex.message
            msgFull = msg+'\n'+''.join(traceback.format_exc())
            msgFull+= "CODE:\n"+code
            if tealGui:
                if teal: teal.popUpErr(tealGui.top, msg, "Action Button Error")
                tealGui.debug(msgFull)
            else:
                if teal: teal.popUpErr(None, msg, "Action Button Error")
                print(msgFull)



from __future__ import division, print_function # confidence high

from stsci.tools import parseinput, fileutil
from astropy.io import fits
import os

def checkFiles(filelist,ivmlist = None):
    """
    - Converts waiver fits sciece and data quality files to MEF format
    - Converts GEIS science and data quality files to MEF format
    - Checks for stis association tables and splits them into single imsets
    - Removes files with EXPTIME=0 and the corresponding ivm files
    - Removes files with NGOODPIX == 0 (to exclude saturated images)
    - Removes files with missing PA_V3 keyword

    The list of science files should match the list of ivm files at the end.
    """

    newfilelist, ivmlist = checkFITSFormat(filelist, ivmlist)

    # check for STIS association files. This must be done before
    # the other checks in order to handle correctly stis
    # assoc files
    #if fits.getval(newfilelist[0], 'INSTRUME') == 'STIS':
    newfilelist, ivmlist = checkStisFiles(newfilelist, ivmlist)
    if newfilelist == []:
        return [], []
    removed_expt_files = check_exptime(newfilelist)

    newfilelist, ivmlist = update_input(newfilelist, ivmlist, removed_expt_files)
    if newfilelist == []:
        return [], []
    removed_ngood_files = checkNGOODPIX(newfilelist)
    newfilelist, ivmlist = update_input(newfilelist, ivmlist, removed_ngood_files)
    if newfilelist == []:
        return [], []

    removed_pav3_files = checkPA_V3(newfilelist)
    newfilelist, ivmlist = update_input(newfilelist, ivmlist, removed_pav3_files)

    newfilelist, ivmlist = update_input(newfilelist, ivmlist,[])

    if newfilelist == []:
        return [], []

    return newfilelist, ivmlist

def checkFITSFormat(filelist, ivmlist=None):
    """
    This code will check whether or not files are GEIS or WAIVER FITS and
    convert them to MEF if found. It also keeps the IVMLIST consistent with
    the input filelist, in the case that some inputs get dropped during
    the check/conversion.
    """
    if ivmlist is None:
        ivmlist = [None for l in filelist]

    sci_ivm = list(zip(filelist, ivmlist))

    removed_files, translated_names, newivmlist = convert2fits(sci_ivm)
    newfilelist, ivmlist = update_input(filelist, ivmlist, removed_files)

    if newfilelist == [] and translated_names == []:
        return [], []

    elif translated_names != []:
        newfilelist.extend(translated_names)
        ivmlist.extend(newivmlist)

    return newfilelist, ivmlist


def checkStisFiles(filelist, ivmlist=None):
    newflist = []
    newilist = []
    removed_files = []
    assoc_files = []
    assoc_ilist = []

    if len(filelist) != len(ivmlist):
        errormsg = "Input file list and ivm list have different lenghts\n"
        errormsg += "Quitting ...\n"
        raise ValueError(errormsg)

    for t in zip(filelist, ivmlist):

        if fits.getval(t[0], 'INSTRUME') != 'STIS':
            newflist.append(t[0])
            newilist.append(t[1])
            continue
        if isSTISSpectroscopic(t[0]):
            removed_files.append(t[0])
            continue
        sci_count = stisObsCount(t[0])
        if sci_count >1:
            newfilenames = splitStis(t[0], sci_count)
            assoc_files.extend(newfilenames)
            removed_files.append(t[0])
            if (isinstance(t[1], tuple) and t[1][0] is not None) or \
               (not isinstance(t[1], tuple) and t[1] is not None):
                print('Does not handle STIS IVM files and STIS association files\n')
            else:
                asn_ivmlist = list(zip(sci_count * [None], newfilenames))
                assoc_ilist.extend(asn_ivmlist)
        elif sci_count == 1:
            newflist.append(t[0])
            newilist.append(t[1])
        else:
            errormsg = "No valid 'SCI extension in STIS file\n"
            raise ValueError(errormsg)

        stisExt2PrimKw([t[0]])

    newflist.extend(assoc_files)
    newilist.extend(assoc_ilist)
    return newflist, newilist

def check_exptime(filelist):
    """
    Removes files with EXPTIME==0 from filelist.
    """
    removed_files = []
    for f in filelist:
        try:
            exptime = fileutil.getHeader(f+'[sci,1]')['EXPTIME']
        except KeyError:
            removed_files.append(f)
            print("Warning:  There are files without keyword EXPTIME")
            continue
        if exptime <= 0:
            removed_files.append(f)
            print("Warning:  There are files with zero exposure time: keyword EXPTIME = 0.0")

    if removed_files != []:
        print("Warning:  Removing the following files from input list")
        for f in removed_files:
            print('\t',f)
    return removed_files

def checkNGOODPIX(filelist):
    """
    Only for ACS, WFC3 and STIS, check NGOODPIX
    If all pixels are 'bad' on all chips, exclude this image
    from further processing.
    Similar checks requiring comparing 'driz_sep_bits' against
    WFPC2 c1f.fits arrays and NICMOS DQ arrays will need to be
    done separately (and later).
    """
    removed_files = []
    supported_instruments = ['ACS','STIS','WFC3']
    for inputfile in filelist:
        if fileutil.getKeyword(inputfile,'instrume') in supported_instruments:
            file = fits.open(inputfile)
            ngood = 0
            for extn in file:
                if 'EXTNAME' in extn.header and extn.header['EXTNAME'] == 'SCI':
                    ngood += extn.header['NGOODPIX']
            file.close()

            if (ngood == 0):
                removed_files.append(inputfile)

    if removed_files != []:
        print("Warning:  Files without valid pixels detected: keyword NGOODPIX = 0.0")
        print("Warning:  Removing the following files from input list")
        for f in removed_files:
            print('\t',f)

    return removed_files

def update_input(filelist, ivmlist=None, removed_files=None):
    """
    Removes files flagged to be removed from the input filelist.
    Removes the corresponding ivm files if present.
    """
    newfilelist = []

    if removed_files == []:
        return filelist, ivmlist
    else:
        sci_ivm = list(zip(filelist, ivmlist))
        for f in removed_files:
            result = [sci_ivm.remove(t) for t in sci_ivm if t[0] == f ]
        ivmlist = [el[1] for el in sci_ivm]
        newfilelist = [el[0] for el in sci_ivm]
        return newfilelist, ivmlist


def stisObsCount(input):
    """
    Input: A stis multiextension file
    Output: Number of stis science extensions in input
    """
    count = 0
    f = fits.open(input)
    for ext in f:
        if 'extname' in ext.header:
            if (ext.header['extname'].upper() == 'SCI'):
                count += 1
    f.close()
    return count

def splitStis(stisfile, sci_count):
    """
    :Purpose: Split a STIS association file into multiple imset MEF files.

    Split the corresponding spt file if present into single spt files.
    If an spt file can't be split or is missing a Warning is printed.

    Returns
    -------
    names: list
        a list with the names of the new flt files.

    """
    newfiles = []

    f = fits.open(stisfile)
    hdu0 = f[0].copy()


    for count in range(1,sci_count+1):
        fitsobj = fits.HDUList()
        fitsobj.append(hdu0)
        hdu = f[('sci',count)].copy()
        fitsobj.append(hdu)
        rootname = hdu.header['EXPNAME']
        newfilename = fileutil.buildNewRootname(rootname, extn='_flt.fits')
        try:
            # Verify error array exists
            if f[('err', count)].data is None:
                raise ValueError
            # Verify dq array exists
            if f[('dq', count)].data is None:
                raise ValueError
            # Copy the err extension
            hdu = f[('err',count)].copy()
            fitsobj.append(hdu)
            # Copy the dq extension
            hdu = f[('dq',count)].copy()
            fitsobj.append(hdu)
            fitsobj[1].header['EXTVER'] = 1
            fitsobj[2].header['EXTVER'] = 1
            fitsobj[3].header['EXTVER'] = 1
        except ValueError:
            print('\nWarning:')
            print('Extension version %d of the input file %s does not' %(count, stisfile))
            print('contain all required image extensions. Each must contain')
            print('populates SCI, ERR and DQ arrays.')

            continue


        # Determine if the file you wish to create already exists on the disk.
        # If the file does exist, replace it.
        if (os.path.exists(newfilename)):
            os.remove(newfilename)
            print("       Replacing "+newfilename+"...")

            # Write out the new file
        fitsobj.writeto(newfilename)
        newfiles.append(newfilename)
    f.close()

    sptfilename = fileutil.buildNewRootname(stisfile, extn='_spt.fits')
    try:
        sptfile = fits.open(sptfilename)
    except IOError:
        print('SPT file not found %s \n' % sptfilename)
        return newfiles

    if sptfile:
        hdu0 = sptfile[0].copy()
        try:
            for count in range(1,sci_count+1):
                fitsobj = fits.HDUList()
                fitsobj.append(hdu0)
                hdu = sptfile[count].copy()
                fitsobj.append(hdu)
                rootname = hdu.header['EXPNAME']
                newfilename = fileutil.buildNewRootname(rootname, extn='_spt.fits')
                fitsobj[1].header['EXTVER'] = 1
                if (os.path.exists(newfilename)):
                    os.remove(newfilename)
                    print("       Replacing "+newfilename+"...")

                # Write out the new file
                fitsobj.writeto(newfilename)
        except:
            print("Warning: Unable to split spt file %s " % sptfilename)
        sptfile.close()

    return newfiles

def stisExt2PrimKw(stisfiles):
    """
        Several kw which are usuall yin the primary header
        are in the extension header for STIS. They are copied to
        the primary header for convenience.
        List if kw:
        'DATE-OBS', 'EXPEND', 'EXPSTART', 'EXPTIME'
    """

    kw_list = ['DATE-OBS', 'EXPEND', 'EXPSTART', 'EXPTIME']

    for sfile in stisfiles:
        d = {}
        for k in kw_list:
            d[k] = fits.getval(sfile, k, ext=1)

        for item in d.items():
            fits.setval(sfile, item[0], value=item[1], comment='Copied from extension header')


def isSTISSpectroscopic(fname):

    if fits.getval(fname, 'OBSTYPE') == 'SPECTROSCOPIC':
        print("Warning:  STIS spectroscopic files detected")
        print("Warning:  Removing %s from input list" % fname)
        return True
    else:
        return False

def checkPA_V3(fnames):
    removed_files = []
    for f in fnames:
        try:
            pav3 = fits.getval(f, 'PA_V3')
        except KeyError:
            rootname = fits.getval(f, 'ROOTNAME')
            sptfile = rootname+'_spt.fits'
            if fileutil.findFile(sptfile):
                try:
                    pav3 = fits.getval(sptfile, 'PA_V3')
                except KeyError:
                    print("Warning:  Files without keyword PA_V3 detected")
                    removed_files.append(f)
                fits.setval(f, 'PA_V3', value=pav3)
            else:
                print("Warning:  Files without keyword PA_V3 detected")
                removed_files.append(f)
    if removed_files != []:
        print("Warning:  Removing the following files from input list")
        for f in removed_files:
            print('\t',f)
    return removed_files

def convert2fits(sci_ivm):
    """
    Checks if a file is in WAIVER of GEIS format and converts it to MEF
    """
    removed_files = []
    translated_names = []
    newivmlist = []

    for file in sci_ivm:
        #find out what the input is
        # if science file is not found on disk, add it to removed_files for removal
        try:
            imgfits,imgtype = fileutil.isFits(file[0])
        except IOError:
            print("Warning:  File %s could not be found" %file[0])
            print("Warning:  Removing file %s from input list" %file[0])
            removed_files.append(file[0])
            continue

        # Check for existence of waiver FITS input, and quit if found.
        # Or should we print a warning and continue but not use that file
        if imgfits and imgtype == 'waiver':
            newfilename = waiver2mef(file[0], convert_dq=True)
            if newfilename is None:
                print("Removing file %s from input list - could not convert WAIVER format to MEF\n" %file[0])
                removed_files.append(file[0])
            else:
                removed_files.append(file[0])
                translated_names.append(newfilename)
                newivmlist.append(file[1])

        # If a GEIS image is provided as input, create a new MEF file with
        # a name generated using 'buildFITSName()'
        # Convert the corresponding data quality file if present
        if not imgfits:
            newfilename = geis2mef(file[0], convert_dq=True)
            if newfilename is None:
                print("Removing file %s from input list - could not convert GEIS format to MEF\n" %file[0])
                removed_files.append(file[0])
            else:
                removed_files.append(file[0])
                translated_names.append(newfilename)
                newivmlist.append(file[1])

    return removed_files, translated_names, newivmlist

def waiver2mef(sciname, newname=None, convert_dq=True):
    """
    Converts a GEIS science file and its corresponding
    data quality file (if present) to MEF format
    Writes out both files to disk.
    Returns the new name of the science image.
    """

    def convert(file):
        newfilename = fileutil.buildNewRootname(file, extn='_c0h.fits')
        try:
            newimage = fileutil.openImage(file,writefits=True,
                                          fitsname=newfilename,clobber=True)
            del newimage
            return newfilename
        except IOError:
            print('Warning: File %s could not be found' % file)
            return None

    newsciname = convert(sciname)
    if convert_dq:
        dq_name = convert(fileutil.buildNewRootname(sciname, extn='_c1h.fits'))

    return newsciname



def geis2mef(sciname, convert_dq=True):
    """
    Converts a GEIS science file and its corresponding
    data quality file (if present) to MEF format
    Writes out both files to disk.
    Returns the new name of the science image.
    """

    def convert(file):
        newfilename = fileutil.buildFITSName(file)
        try:
            newimage = fileutil.openImage(file,writefits=True,
                fitsname=newfilename, clobber=True)
            del newimage
            return newfilename
        except IOError:
            print('Warning: File %s could not be found' % file)
            return None

    newsciname = convert(sciname)
    if convert_dq:
        dq_name = convert(sciname.split('.')[0] + '.c1h')

    return newsciname

def countInput(input):
    files = parseinput.parseinput(input)
    count = len(files[0])
    for f in files[0]:
        if fileutil.isFits(f)[0]:
            try:
                ins = fits.getval(f, 'INSTRUME')
            except: # allow odd fits files; do not stop the count
                ins = None
            if ins == 'STIS':
                count += (stisObsCount(f)-1)
    return count
